"""Utils for sweeper management."""

from collections import defaultdict

from keysight import qcs

from qibolab._core.identifier import ChannelId
from qibolab._core.pulses import PulseId
from qibolab._core.sweeper import ParallelSweepers, Parameter

NS_TO_S = 1e-9
SUPPORTED_CHANNEL_SWEEPERS = [Parameter.frequency]
SUPPORTED_PULSE_SWEEPERS = [
    Parameter.amplitude,
    Parameter.duration,
    Parameter.relative_phase,
]


def process_sweepers(
    sweepers: list[ParallelSweepers], probe_channel_ids: set[ChannelId]
):
    """Processes Qibocal sweepers into QCS sweepers. Currently nested hardware
    sweepers are not supported, so they will default to software sweeping.

    Arguments:
        sweepers (list[ParallelSweepers]): Array of array of sweepers.
        probe_channel_ids (list[int]): Array of channel IDs for probe channels.

    Returns:
        hardware_sweepers (list[tuple[list[qcs.Array], list[qcs.Scalar]]]): Array of hardware-based QCS sweepers.
        software_sweepers (list[tuple[list[qcs.Array], list[qcs.Scalar]]]): Array of software-based QCS sweepers.
        sweeper_channel_map (dict[ChannelId, qcs.Scalar]): Map of channel ID to frequency to be swept.
        sweeper_pulse_map (defaultdict[PulseId, dict[str, qcs.Scalar]]): Map of pulse ID to map of parameter
        to be swept and corresponding QCS variable.
    """
    hardware_sweepers: list[tuple[list[qcs.Array], list[qcs.Scalar]]] = []
    software_sweepers: list[tuple[list[qcs.Array], list[qcs.Scalar]]] = []

    # Mapper for pulses that are controlled by a sweeper and the parameter to be swept
    sweeper_pulse_map: defaultdict[PulseId, dict[str, qcs.Scalar]] = defaultdict(dict)
    # Mapper for channels with frequency controlled by a sweeper
    sweeper_channel_map: dict[ChannelId, qcs.Scalar] = {}

    for idx, parallel_sweeper in enumerate(reversed(sweepers)):
        sweep_values: list[qcs.Array] = []
        sweep_variables: list[qcs.Variable] = []
        # Currently nested hardware sweeping is not supported
        hardware_sweeping = len(hardware_sweepers + software_sweepers) == 0

        for idx2, sweeper in enumerate(parallel_sweeper):
            qcs_variable = qcs.Scalar(
                name=f"V{idx}_{idx2}", value=sweeper.values[0], dtype=float
            )

            if sweeper.parameter in SUPPORTED_CHANNEL_SWEEPERS:
                sweeper_channel_map.update(
                    {channel_id: qcs_variable for channel_id in sweeper.channels}
                )
                # Readout frequency is not supported with hardware sweeping
                if not probe_channel_ids.isdisjoint(sweeper.channels):
                    hardware_sweeping = False
            elif sweeper.parameter in SUPPORTED_PULSE_SWEEPERS:
                # Duration is not supported with hardware sweeping
                if sweeper.parameter is Parameter.duration:
                    hardware_sweeping = False

                for pulse in sweeper.pulses:
                    sweeper_pulse_map[pulse.id][sweeper.parameter.name] = qcs_variable
            else:
                raise ValueError(
                    "Sweeper parameter not supported", sweeper.parameter.name
                )

            sweep_variables.append(qcs_variable)
            sweep_values.append(
                qcs.Array(
                    name=f"A{idx}_{idx2}",
                    value=(
                        sweeper.values * NS_TO_S
                        if sweeper.parameter is Parameter.duration
                        else sweeper.values
                    ),
                    dtype=float,
                )
            )
        if hardware_sweeping:
            hardware_sweepers.append((sweep_values, sweep_variables))
        else:
            software_sweepers.append((sweep_values, sweep_variables))

    return (
        hardware_sweepers,
        software_sweepers,
        sweeper_channel_map,
        sweeper_pulse_map,
    )



from collections.abc import Iterable
from typing import Annotated, Optional, Union

from pydantic import UUID4, AfterValidator

from qibolab._core.pulses import Pulse, PulseId, PulseLike, Readout
from qibolab._core.serialize import ArrayList, Model
from qibolab._core.sweeper import Sweeper

__all__ = []

ComponentId = tuple[UUID4, int]
WaveformIndices = dict[ComponentId, tuple[int, int]]


class Waveform(Model):
    data: Annotated[ArrayList, AfterValidator(lambda a: a.astype(float))]
    index: int


class WaveformSpec(Model):
    waveform: Waveform
    duration: int


Waveforms = dict[ComponentId, Waveform]


def _pulse(event: Union[Pulse, Readout]) -> Pulse:
    return event.probe if isinstance(event, Readout) else event


def waveforms(
    sequence: Iterable[PulseLike],
    sampling_rate: float,
    amplitude_swept: set[PulseId],
    duration_swept: dict[PulseLike, Sweeper],
) -> dict[ComponentId, WaveformSpec]:
    def _waveform(
        pulse: Pulse, component: str, duration: Optional[float] = None
    ) -> WaveformSpec:
        duration_ = pulse.duration if duration is None else duration
        update = {"duration": duration_} | (
            {"amplitude": 1.0} if pulse.id in amplitude_swept else {}
        )
        return WaveformSpec(
            waveform=Waveform(
                data=getattr(pulse.model_copy(update=update), component)(sampling_rate),
                index=0,
            ),
            duration=int(duration_),
        )

    indexless = {
        k: v
        for d in (
            {
                (pulse.id, 0): _waveform(pulse, "i"),
                (pulse.id, 1): _waveform(pulse, "q"),
            }
            for pulse in (
                _pulse(event)
                for event in sequence
                if isinstance(event, (Pulse, Readout))
            )
            if pulse not in duration_swept
        )
        for k, v in d.items()
    } | {
        k: v
        for d in (
            {
                (pulse.id, 2 * i): _waveform(pulse, "i", duration),
                (pulse.id, 2 * i + 1): _waveform(pulse, "q", duration),
            }
            for pulse, sweep in (
                (_pulse(event), duration_swept[event])
                for event in duration_swept
                if isinstance(event, (Pulse, Readout))
            )
            for i, duration in (
                (i, sweep.irange[0] + sweep.irange[2] * i) for i in range(len(sweep))
            )
        )
        for k, v in d.items()
    }

    return {
        k: WaveformSpec(
            waveform=Waveform(data=v.waveform.data, index=i), duration=v.duration
        )
        for i, (k, v) in enumerate(indexless.items())
    }


def waveform_indices(waveforms: dict[ComponentId, WaveformSpec]) -> WaveformIndices:
    return {k: (w.waveform.index, w.duration) for k, w in waveforms.items()}



from typing import Literal, Union

import numpy as np

from qibolab._core.components import (
    AcquisitionConfig,
    DcConfig,
    OscillatorConfig,
)
from qibolab._core.components.filters import ExponentialFilter

__all__ = [
    "OpxOutputConfig",
    "QmAcquisitionConfig",
    "QmConfigs",
    "OctaveOscillatorConfig",
    "MwFemOscillatorConfig",
]

OctaveOutputModes = Literal[
    "always_on", "always_off", "triggered", "triggered_reversed"
]

DEFAULT_SAMPLING_RATE = 1e9

DEFAULT_FEEDFORWARD_MAX = 2 - 2**-16
"""Maximum feedforward tap value"""
DEFAULT_FEEDBACK_MAX = 1 - 2**-20
"""Maximum feedback tap value"""


def normalize_feedforward(taps: list[float], threshold: float) -> list[float]:
    """Feedforward coefficient normalization required by QM."""
    scale = np.max(np.abs(taps) / threshold, initial=1)
    return (np.array(taps) / scale).tolist()


def normalize_feedback(taps: list[float], threshold: float) -> list[float]:
    """Feedback coefficient normalization required by QM."""
    new_taps = np.clip(taps, -threshold, threshold)
    return new_taps.tolist()


class OpxOutputConfig(DcConfig):
    """DC channel config using QM OPX+."""

    kind: Literal["opx-output"] = "opx-output"

    offset: float = 0.0
    """DC offset to be applied in V.

    Possible values are -0.5V to 0.5V.
    """
    output_mode: Literal["direct", "amplified"] = "direct"
    sampling_rate: float = DEFAULT_SAMPLING_RATE
    upsampling_mode: Literal["mw", "pulse"] = "mw"
    feedback_max: float = DEFAULT_FEEDBACK_MAX
    feedforward_max: float = DEFAULT_FEEDFORWARD_MAX

    @property
    def filter(self):
        feedback_filters = [
            -i.feedback[1] for i in self.filters if isinstance(i, ExponentialFilter)
        ]
        return {
            "feedback": normalize_feedback(feedback_filters, self.feedback_max)
            if len(feedback_filters) > 0
            else [],
            "feedforward": normalize_feedforward(self.feedforward, self.feedforward_max)
            if len(self.feedforward) > 0
            else [],
        }


class OctaveOscillatorConfig(OscillatorConfig):
    """Oscillator confing that allows switching the output mode."""

    kind: Literal["octave-oscillator"] = "octave-oscillator"

    output_mode: OctaveOutputModes = "triggered"


class QmAcquisitionConfig(AcquisitionConfig):
    """Acquisition config for QM OPX+."""

    kind: Literal["qm-acquisition"] = "qm-acquisition"

    gain: int = 0
    """Input gain in dB.

    Possible values are -12dB to 20dB in steps of 1dB.
    """
    offset: float = 0.0
    """Constant voltage to be applied on the input."""


class MwFemOscillatorConfig(OscillatorConfig):
    """Output config for OPX1000 MW-FEM ports.

    For more information see
    https://docs.quantum-machines.co/latest/docs/Guides/opx1000_fems/?h=upsampl#microwave-fem-mw-fem
    """

    kind: Literal["mw-fem-oscillator"] = "mw-fem-oscillator"

    power: int = -11
    """This corresponds to the ``full_scale_power_dbm`` setting."""
    upconverter: int = 1
    band: int = 2
    sampling_rate: float = DEFAULT_SAMPLING_RATE


QmConfigs = Union[
    OpxOutputConfig,
    OctaveOscillatorConfig,
    QmAcquisitionConfig,
    MwFemOscillatorConfig,
]



"""Qibolab driver for Keysight QCS instrument set."""

from collections import defaultdict
from functools import reduce
from typing import ClassVar

import numpy as np
from keysight import qcs

from qibolab._core.components import AcquisitionChannel, Config, DcChannel, IqChannel
from qibolab._core.execution_parameters import AveragingMode, ExecutionParameters
from qibolab._core.identifier import ChannelId, Result
from qibolab._core.instruments.abstract import Controller
from qibolab._core.pulses import PulseId
from qibolab._core.sequence import InputOps, PulseSequence
from qibolab._core.sweeper import ParallelSweepers

from .pulse import (
    process_acquisition_channel_pulses,
    process_dc_channel_pulses,
    process_iq_channel_pulses,
)
from .results import fetch_result, parse_result
from .sweep import process_sweepers

NS_TO_S = 1e-9

__all__ = ["KeysightQCS"]


def sweeper_reducer(
    program: qcs.Program, sweepers: tuple[list[qcs.Array], list[qcs.Scalar]]
):
    """Helper method to unpack the QCS sweep parameters when processing sweepers."""
    return program.sweep(*sweepers)


class KeysightQCS(Controller):
    """Driver for interacting with QCS controller server."""

    bounds: str = "qcs/bounds"

    qcs_channel_map: qcs.ChannelMapper
    """Map of QCS virtual channels to QCS physical channels."""
    virtual_channel_map: dict[ChannelId, qcs.Channels]
    """Map of Qibolab channel IDs to QCS virtual channels."""
    sampling_rate: ClassVar[float] = (
        qcs.SAMPLE_RATES[qcs.InstrumentEnum.M5300AWG] * NS_TO_S
    )

    def connect(self):
        self.backend = qcs.HclBackend(self.qcs_channel_map, hw_demod=True)
        self.backend.is_system_ready()

    def create_program(
        self,
        sequence: PulseSequence,
        configs: dict[str, Config],
        sweepers: list[ParallelSweepers],
        num_shots: int,
    ) -> tuple[qcs.Program, list[tuple[int, int]]]:
        # SWEEPER MANAGEMENT
        probe_channel_ids = {
            chan.probe
            for chan in self.channels.values()
            if isinstance(chan, AcquisitionChannel)
        }

        (
            hardware_sweepers,
            software_sweepers,
            sweeper_channel_map,
            sweeper_pulse_map,
        ) = process_sweepers(sweepers, probe_channel_ids)
        # Here we are telling the program to run hardware sweepers first, then software sweepers
        # It is essential that we match the original sweeper order to the modified sweeper order
        # to reconcile the results at the end
        program = reduce(
            sweeper_reducer,
            software_sweepers,
            reduce(sweeper_reducer, hardware_sweepers, qcs.Program()).n_shots(
                num_shots
            ),
        )

        # WAVEFORM COMPILATION
        # Iterate over channels and convert qubit pulses to QCS waveforms
        for channel_id in sequence.channels:
            channel = self.channels[channel_id]
            virtual_channel = self.virtual_channel_map[channel_id]

            if isinstance(channel, AcquisitionChannel):
                probe_channel_id = channel.probe
                classifier_reference = configs[channel_id].state_iq_values
                process_acquisition_channel_pulses(
                    program=program,
                    pulses=sequence.channel(channel_id),
                    frequency=sweeper_channel_map.get(
                        probe_channel_id, configs[probe_channel_id].frequency
                    ),
                    virtual_channel=virtual_channel,
                    probe_virtual_channel=self.virtual_channel_map[probe_channel_id],
                    sweeper_pulse_map=sweeper_pulse_map,
                    classifier=(
                        None
                        if classifier_reference is None
                        else qcs.MinimumDistanceClassifier(classifier_reference)
                    ),
                )

            elif isinstance(channel, IqChannel):
                process_iq_channel_pulses(
                    program=program,
                    pulses=sequence.channel(channel_id),
                    frequency=sweeper_channel_map.get(
                        channel_id, configs[channel_id].frequency
                    ),
                    virtual_channel=virtual_channel,
                    sweeper_pulse_map=sweeper_pulse_map,
                )

            elif isinstance(channel, DcChannel):
                process_dc_channel_pulses(
                    program=program,
                    pulses=sequence.channel(channel_id),
                    virtual_channel=virtual_channel,
                    sweeper_pulse_map=sweeper_pulse_map,
                )

        return program

    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ) -> dict[int, Result]:
        if options.relaxation_time is not None:
            self.backend._init_time = int(options.relaxation_time)

        ret: dict[PulseId, np.ndarray] = {}
        for sequence in sequences:
            results = self.backend.apply(
                self.create_program(
                    sequence.align_to_delays(), configs, sweepers, options.nshots
                )
            ).results
            acquisition_map: defaultdict[qcs.Channels, list[InputOps]] = defaultdict(
                list
            )

            for channel_id, input_op in sequence.acquisitions:
                channel = self.virtual_channel_map[channel_id]
                acquisition_map[channel].append(input_op)

            averaging = options.averaging_mode is not AveragingMode.SINGLESHOT
            for channel, input_ops in acquisition_map.items():
                raw = fetch_result(
                    results=results,
                    channel=channel,
                    acquisition_type=options.acquisition_type,
                    averaging=averaging,
                )

                for result, input_op in zip(raw.values(), input_ops):
                    ret[input_op.id] = parse_result(result, options)

        return ret

    def disconnect(self):
        pass



import hashlib
from typing import List

import numpy as np
import numpy.typing as npt

from .abstract import Extractor

__all__ = ["ShaExtractor"]


class ShaExtractor(Extractor):
    """Extractor based on the SHA-256 hash algorithm."""

    def num_raw_samples(self, n: int) -> int:
        return 22 * (n // 4 + 1)

    def extract(self, raw: List[int]) -> npt.NDArray:
        extracted = []
        for i in range(len(raw) // 22):
            stream = "".join(
                format(sample, "012b") for sample in raw[22 * i : 22 * (i + 1)]
            )
            hash = hashlib.sha256(stream.encode("utf-8")).hexdigest()
            sha_bin = bin(int(hash, 16))[2:].zfill(256)
            for j in range(4):
                # Convert 53-bit chunk to integer
                uniform_int = int(sha_bin[53 * j : 53 * (j + 1)], 2)
                extracted.append(uniform_int / (2**53 - 1))
        return np.array(extracted)



"""Utils for result management."""

import numpy as np
from keysight import qcs

from qibolab._core.execution_parameters import (
    AcquisitionType,
    AveragingMode,
    ExecutionParameters,
)


def fetch_result(
    results: qcs.Results,
    channel: qcs.Channels,
    acquisition_type: AcquisitionType,
    averaging: bool,
) -> dict[qcs.Channels, np.ndarray]:
    """Processes the QCS result object to return the appropiate results.

    Arguments:
        results (qcs.Results): QCS result object.
        channel (qcs.Channels): Virtual acquisition channel to obtain results from.
        acquisition_type (AcquisitionType): Acquisition type to be used.
        averaging (bool): Flag for averaging results
        sweeper_swaps_required (list[tuple[int, int]]): Array of axes pairs corresponding to swapped sweepers.

    Returns:
        raw (dict[qcs.Channels, np.ndarray]): Map of virtual channel to acquisition results.
    """
    if acquisition_type is AcquisitionType.RAW:
        raw = results.get_trace(channel, averaging)
    elif acquisition_type is AcquisitionType.INTEGRATION:
        raw = results.get_iq(channel, averaging)
    elif acquisition_type is AcquisitionType.DISCRIMINATION:
        raw = results.get_classified(channel, averaging)
    else:
        raise ValueError("Acquisition type unrecognized")
    return raw


def parse_result(result: np.ndarray, options: ExecutionParameters) -> np.ndarray:
    """Parses resulting numpy array into Qibolab expected array shape.

    Arguments:
        result (np.ndarray): Result array from QCS.
        options (ExecutionParameters): Execution settings.
        sweepers (list[ParallelSweepers]): Array of array of sweepers.

    Returns:
        parsed_result (np.ndarray): Parsed numpy array.
    """
    if (
        options.averaging_mode is not AveragingMode.SINGLESHOT
        and options.acquisition_type is not AcquisitionType.INTEGRATION
    ):
        return result
    # For single shot, qibolab expects result format (nshots, ...)
    # QCS returns (..., nshots), so we need to shuffle the arrays
    if options.averaging_mode is AveragingMode.SINGLESHOT:
        result = np.moveaxis(result, -1, 0)
    # For IQ data, QCS returns complex results
    if options.acquisition_type is not AcquisitionType.INTEGRATION:
        return result
    return np.moveaxis(np.stack([np.real(result), np.imag(result)]), 0, -1)



from abc import ABC, abstractmethod
from typing import List

import numpy.typing as npt

from ....serialize import Model

__all__ = ["Extractor"]


class Extractor(Model, ABC):
    @abstractmethod
    def num_raw_samples(self, n: int) -> int:
        """Number of raw QRNG samples that are needed to reach the required random floats.

        Args:
            n (int): Number of required random floats.
        """

    @abstractmethod
    def extract(self, raw: List[int]) -> npt.NDArray:
        """Extract uniformly distributed integers from the device samples."""



from enum import Enum, auto
from functools import cache
from typing import Any, Collection, Optional

import numpy as np
import numpy.typing as npt
from pydantic import model_validator

from .components.configs import OscillatorConfig
from .identifier import ChannelId
from .pulses import PulseLike, VirtualZ
from .serialize import Model

__all__ = ["Parameter", "ParallelSweepers", "Sweeper"]

_PULSE = "pulse"
_CHANNEL = "channel"


class Parameter(Enum):
    """Sweeping parameters."""

    frequency = (auto(), _CHANNEL)
    amplitude = (auto(), _PULSE)
    duration = (auto(), _PULSE)
    duration_interpolated = (auto(), _PULSE)
    relative_phase = (auto(), _PULSE)
    phase = (auto(), _PULSE)
    offset = (auto(), _CHANNEL)

    @classmethod
    @cache
    def channels(cls) -> set["Parameter"]:
        """Set of parameters to be swept on the channel."""
        return {p for p in cls if p.value[1] == _CHANNEL}


_Field = tuple[Any, str]


def _alternative_fields(a: _Field, b: _Field):
    if (a[0] is None) == (b[0] is None):
        raise ValueError(
            f"Either '{a[1]}' or '{b[1]}' needs to be provided, and only one of them."
        )


Range = tuple[float, float, float]


class Sweeper(Model):
    """Data structure for Sweeper object.

    This object is passed as an argument to the method :func:`qibolab.Platform.execute`
    which enables the user to sweep a specific parameter for one or more pulses. For information on how to
    perform sweeps see :func:`qibolab.Platform.execute`.

    Example:
        .. testcode::

            import numpy as np
            from qibolab import Parameter, PulseSequence, Sweeper
            from qibolab.instruments.dummy import create_dummy


            platform = create_dummy()
            qubit = platform.qubits[0]
            natives = platform.natives.single_qubit[0]
            sequence = natives.MZ.create_sequence()
            parameter_range = np.random.randint(10, size=10)
            sweeper = Sweeper(
                parameter=Parameter.frequency, values=parameter_range, channels=[qubit.probe]
            )
            platform.execute([sequence], [[sweeper]])
    """

    parameter: Parameter
    """Parameter to be swept."""
    values: Optional[npt.NDArray] = None
    """Array of parameter values to sweep over."""
    range: Optional[Range] = None
    """Tuple of ``(start, stop, step)``.

    To sweep over the array ``np.arange(start, stop, step)``.
    Can be provided instead of ``values`` for more efficient sweeps on some instruments.
    """
    pulses: Optional[list[PulseLike]] = None
    """List of `qibolab.Pulse` to be swept."""
    channels: Optional[list[ChannelId]] = None
    """List of channel names for which the parameter should be swept."""

    @model_validator(mode="after")
    def check_values(self):
        _alternative_fields((self.pulses, "pulses"), (self.channels, "channels"))

        if self.values is None and self.range is None:
            raise ValueError("Cannot create a sweeper without values or range.")
        if self.pulses is not None and self.parameter in Parameter.channels():
            raise ValueError(
                f"Cannot create a sweeper for {self.parameter} without specifying channels."
            )
        if self.parameter not in Parameter.channels() and (self.channels is not None):
            raise ValueError(
                f"Cannot create a sweeper for {self.parameter} without specifying pulses."
            )
        if self.parameter is Parameter.phase and not all(
            isinstance(pulse, VirtualZ) for pulse in self.pulses
        ):
            raise TypeError("Cannot create a phase sweeper on non-VirtualZ pulses.")

        if self.range is not None:
            object.__setattr__(self, "values", np.arange(*self.range))

        if self.parameter is Parameter.amplitude and max(abs(self.values)) > 1:
            raise ValueError(
                "Amplitude sweeper cannot have absolute values larger than 1."
            )

        return self

    @property
    def irange(self) -> tuple[float, float, float]:
        """Inferred range.

        Always ensure a range, inferring it from :attr:`values` if :attr:`range` is
        not set.
        """
        if self.range is not None:
            return self.range
        assert self.values is not None
        step = self.values[1] - self.values[0]
        return (self.values[0], self.values[-1] + step, step)

    def __len__(self) -> int:
        """Compute number of iterations."""
        if self.values is not None:
            return len(self.values)
        assert self.range is not None
        return int((self.range[1] - self.range[0]) // self.range[2] + 1)

    def __add__(self, value: float) -> "Sweeper":
        """Add value to sweeper ones."""
        return self.model_copy(
            update=(
                {"range": (self.range[0] + value, self.range[1] + value, self.range[2])}
                if self.range is not None
                else {}
            )
            | ({"values": self.values + value} if self.values is not None else {})
        )

    def __sub__(self, value: float) -> "Sweeper":
        """Subtract value from sweeper ones."""
        return self + (-value)

    def __mul__(self, value: float) -> "Sweeper":
        """Multiply value to sweeper ones.

        TODO: deduplicate this and :meth:`__add__`
        """
        return self.model_copy(
            update=(
                {"range": (self.range[0] * value, self.range[1] * value, self.range[2])}
                if self.range is not None
                else {}
            )
            | ({"values": self.values * value} if self.values is not None else {})
        )

    def __truediv__(self, value: float) -> "Sweeper":
        """Divide by value from sweeper ones."""
        return self * (1 / value)


ParallelSweepers = list[Sweeper]
"""Sweepers that should be iterated in parallel."""


def iteration_length(sweepers: ParallelSweepers) -> int:
    """Compute lenght of parallel iteration."""
    return min((len(s) for s in sweepers), default=0)


def swept_pulses(
    sweepers: list[ParallelSweepers],
    parameters: Collection[Parameter] = frozenset(Parameter),
) -> dict[PulseLike, Sweeper]:
    """Associate pulses swept to sweepers.

    Essentially, it produces a reverse index from `sweepers`.

    If `parameters` is passed, it limits the selection to pulses whose parameter swept
    is among those listed. By default, all swept pulses are returned.
    """
    # TODO: this is assuming a pulse is only swept by a single sweeper. Which is not
    # always the case. A list of `Sweeper` objects should be returned instead
    return {
        p: sweep
        for parsweep in sweepers
        for sweep in parsweep
        if sweep.parameter in parameters and sweep.pulses is not None
        for p in sweep.pulses
    }


def swept_channels(sweepers: list[ParallelSweepers]) -> set[ChannelId]:
    """Identify channels involved in a sweeper suite."""
    return {
        channel
        for ps in sweepers
        for sweeper in ps
        for channel in (sweeper.channels if sweeper.channels is not None else [])
    }


def _split_sweeper(sweeper: Sweeper) -> ParallelSweepers:
    return (
        [sweeper.model_copy(update={"pulses": [pulse]}) for pulse in sweeper.pulses]
        if sweeper.pulses is not None
        else [sweeper.model_copy(update={"channels": [ch]}) for ch in sweeper.channels]
    )


def _split_sweepers(sweepers: list[ParallelSweepers]) -> list[ParallelSweepers]:
    return [
        [s for sweep in parsweep for s in _split_sweeper(sweep)]
        for parsweep in sweepers
    ]


def _lo_frequency(lo: Optional[OscillatorConfig]) -> float:
    return lo.frequency if lo is not None else 0.0


def _subtract_lo(
    sweepers: list[ParallelSweepers], los: dict[ChannelId, OscillatorConfig]
) -> list[ParallelSweepers]:
    return [
        [
            (sweep - _lo_frequency(los.get(sweep.channels[0])))
            if sweep.parameter is Parameter.frequency
            else sweep
            for sweep in parsweep
        ]
        for parsweep in sweepers
    ]


def _subtract_offset(
    sweepers: list[ParallelSweepers], offsets: dict[ChannelId, float]
) -> list[ParallelSweepers]:
    return [
        [
            (sweep - offsets.get(sweep.channels[0], 0.0))
            if sweep.parameter is Parameter.offset
            else sweep
            for sweep in parsweep
        ]
        for parsweep in sweepers
    ]


def normalize_sweepers(
    sweepers: list[ParallelSweepers],
    los: Optional[dict[ChannelId, OscillatorConfig]] = None,
    offsets: Optional[dict[ChannelId, float]] = None,
) -> list[ParallelSweepers]:
    sweepers = _split_sweepers(sweepers)
    sweepers = _subtract_lo(sweepers, los) if los is not None else sweepers
    return _subtract_offset(sweepers, offsets) if offsets is not None else sweepers



from typing import Any, Type

import qblox_instruments as qblox
import qcodes.instrument
from qcodes.instrument import Instrument

from .cluster import MockCluster

__all__ = ["install"]

qcodes_find_or_create_instrument = qcodes.instrument.find_or_create_instrument


def find_or_create_instrument(
    instrument_class: Type[Instrument], *args, **kwargs
) -> Any:
    if issubclass(instrument_class, qblox.Cluster):
        return MockCluster(*args, **kwargs)

    return qcodes_find_or_create_instrument(instrument_class, *args, **kwargs)


def install():
    """Replace qblox instruments instantiation with mocked elements."""
    qcodes.instrument.find_or_create_instrument = find_or_create_instrument



from enum import Flag, auto
from typing import Annotated, Any, Literal, cast

from qblox_instruments.qcodes_drivers.module import Module

from qibolab._core.components import Channel, OscillatorConfig
from qibolab._core.components.configs import Configs
from qibolab._core.identifier import ChannelId
from qibolab._core.serialize import Model

from .port import PortAddress

__all__ = []


def los(
    all: dict[ChannelId, str],
    configs: Configs,
    module_channels: list[tuple[ChannelId, PortAddress]],
) -> dict[ChannelId, OscillatorConfig]:
    return {
        id_: cast(OscillatorConfig, configs[lo])
        for id_, lo in all.items()
        if id_ in {ch[0] for ch in module_channels}
    }


class ModuleType(Flag):
    """Module types, used to declaratively restrict configurations scope."""

    QRM = auto()
    """QRM type module."""
    RF = auto()
    """Radio frequency module."""


class ModuleConfig(Model):
    los: dict[str, Any]
    """Local oscillators configurations."""
    # the following attributes are automatically processed and set
    scope_acq_trigger_mode_path0: Annotated[
        Literal["sequencer", "level"], ModuleType.QRM
    ] = "sequencer"
    """Scope trigger mode for path 0.

    .. note::

        Acquisition on external digital trigger currently unsupported.
    """
    scope_acq_trigger_mode_path1: Annotated[
        Literal["sequencer", "level"], ModuleType.QRM
    ] = "sequencer"
    """Scope trigger mode for path 0.

    Cf. :attr:`scope_acq_trigger_mode_path0`.
    """
    in0_att: Annotated[int, ModuleType.QRM] = 0
    """Input attenuation."""

    @classmethod
    def compute(
        cls,
        channels: dict[ChannelId, Channel],
        los: dict[ChannelId, OscillatorConfig],
        qrm: bool,
    ) -> "ModuleConfig":
        los_ = {}
        # set lo frequencies
        for iq, lo in los.items():
            n = PortAddress.from_path(channels[iq].path).ports[0] - 1
            path = f"out{n}_in{n}" if qrm else f"out{n}"
            los_[f"{path}_lo_en"] = True
            los_[f"{path}_lo_freq"] = int(lo.frequency)
            los_[f"out{n}_att"] = int(lo.power)

        return cls(los=los_)

    @staticmethod
    def _set_option(mod: Module, name: str, metadata: list, value: Any) -> None:
        # - avoid configuring not explicitly set values
        # - los configurations have dynamical names, they are handled separately
        if value is None or name == "los":
            return

        flag = [m for m in metadata if isinstance(m, ModuleType)]
        if len(flag) > 0:
            assert len(flag) == 1
            modtype = flag[0]
            if ModuleType.QRM in modtype and not mod.is_qrm_type:
                return
            if ModuleType.RF in modtype and not mod.is_rf_type:
                return

        mod.set(name, value)

    def apply(self, mod: Module) -> None:
        """Configure module-wide settings."""
        # first disable all default sequencer connections
        mod.disconnect_outputs()

        if mod.is_qrm_type:
            # including input ones, if QRM
            mod.disconnect_inputs()

        # only RF modules have LOs configured
        assert mod.is_rf_type or len(self.los) == 0
        for config, value in self.los.items():
            mod.set(config, value)

        # apply all the other configurations
        for name, field in self.model_fields.items():
            self._set_option(mod, name, field.metadata, getattr(self, name))



"""A platform for executing quantum algorithms."""

import logging
import signal
from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal, Optional

from ..components import Config
from ..components.channels import Channel
from ..execution_parameters import ExecutionParameters
from ..identifier import ChannelId, QubitId, QubitPairId, Result
from ..instruments.abstract import Controller
from ..parameters import (
    InstrumentMap,
    NativeGates,
    Parameters,
    QubitMap,
    Settings,
    Update,
    update_configs,
)
from ..pulses import PulseId
from ..qubits import Qubit
from ..sequence import PulseSequence
from ..sweeper import ParallelSweepers
from ..unrolling import Bounds, batch

__all__ = ["Platform"]

PARAMETERS = "parameters.json"

log = logging.getLogger(__name__)


def _channels_map(elements: QubitMap) -> dict[ChannelId, QubitId]:
    """Map channel names to element (qubit or coupler)."""
    return {ch: id for id, el in elements.items() for ch in el.channels}


def _unique_acquisitions(sequences: list[PulseSequence]) -> bool:
    """Check unique acquisition identifiers."""
    ids = []
    for seq in sequences:
        ids += (p.id for _, p in seq.acquisitions)

    return len(ids) == len(set(ids))


@dataclass
class Platform:
    """Platform for controlling quantum devices."""

    name: str
    """Name of the platform."""
    parameters: Parameters
    """..."""
    instruments: InstrumentMap
    """Mapping instrument names to
    :class:`qibolab.instruments.abstract.Instrument` objects."""
    qubits: QubitMap
    """Qubit controllers.

    The mapped objects hold the :class:`qubit.components.channels.Channel` instances
    required to send pulses addressing the desired qubits.
    """
    couplers: QubitMap = field(default_factory=dict)
    """Coupler controllers.

    Fully analogue to :attr:`qubits`. Only the flux channel is expected to be populated
    in the mapped objects.
    """
    resonator_type: Literal["2D", "3D"] = "2D"
    """Type of resonator (2D or 3D) in the used QPU."""
    is_connected: bool = False
    """Flag for whether we are connected to the physical instruments."""

    def __post_init__(self):
        log.info("Loading platform %s", self.name)
        signal.signal(signal.SIGTERM, self.termination_handler)
        signal.signal(signal.SIGINT, self.termination_handler)
        if self.resonator_type is None:
            self.resonator_type = "3D" if self.nqubits == 1 else "2D"

    def __str__(self):
        return self.name

    @property
    def nqubits(self) -> int:
        """Total number of usable qubits in the QPU."""
        return len(self.qubits)

    @property
    def pairs(self) -> list[QubitPairId]:
        """Available pairs in thee platform."""
        return list(self.parameters.native_gates.two_qubit)

    @property
    def ordered_pairs(self):
        """List of qubit pairs that are connected in the QPU."""
        return sorted({tuple(sorted(pair)) for pair in self.pairs})

    @property
    def settings(self) -> Settings:
        """Container with default execution settings."""
        return self.parameters.settings

    @property
    def natives(self) -> NativeGates:
        """Native gates containers."""
        return self.parameters.native_gates

    @property
    def sampling_rate(self):
        """Sampling rate of control electronics in giga samples per second
        (GSps)."""
        for instrument in self.instruments.values():
            if isinstance(instrument, Controller):
                return instrument.sampling_rate

    @property
    def components(self) -> set[str]:
        """Names of all components available in the platform."""
        return set(self.parameters.configs.keys())

    @property
    def channels(self) -> dict[ChannelId, Channel]:
        """Channels in the platform."""
        return {
            id: ch
            for instr in self.instruments.values()
            if isinstance(instr, Controller)
            for id, ch in instr.channels.items()
        }

    @property
    def qubit_channels(self) -> dict[ChannelId, QubitId]:
        """Channel to qubit map."""
        return _channels_map(self.qubits)

    @property
    def coupler_channels(self):
        """Channel to coupler map."""
        return _channels_map(self.couplers)

    def config(self, name: str) -> Config:
        """Returns configuration of given component."""
        # pylint: disable=unsubscriptable-object
        return self.parameters.configs[name]

    def update(self, update: Update):
        """Update platform's parameters."""
        self.parameters = self.parameters.replace(update)

    def connect(self):
        """Connect to all instruments."""
        if not self.is_connected:
            for name, instrument in self.instruments.items():
                try:
                    instrument.connect()
                except Exception as exception:
                    raise RuntimeError(
                        f"Cannot establish connection to instrument {name}. Error captured: '{exception}'",
                    )
        self.is_connected = True

    def disconnect(self):
        """Disconnects from instruments."""
        if self.is_connected:
            for instrument in self.instruments.values():
                instrument.disconnect()
        self.is_connected = False

    def termination_handler(self, signum, frame):
        self.disconnect()
        raise RuntimeError(
            f"Platform {self.name} disconnected because job was cancelled. Signal type: {signum}."
        )

    @property
    def _controller(self):
        """Identify controller instrument.

        Used for splitting the unrolled sequences to batches.

        This method does not support platforms with more than one
        controller instruments.
        """
        controllers = [
            instr
            for instr in self.instruments.values()
            if isinstance(instr, Controller)
        ]
        assert len(controllers) == 1
        return controllers[0]

    def _execute(
        self,
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        configs: dict[str, Config],
        sweepers: list[ParallelSweepers],
    ) -> dict[PulseId, Result]:
        """Execute sequences on the controllers."""
        result = {}

        for instrument in self.instruments.values():
            if isinstance(instrument, Controller):
                new_result = instrument.play(configs, sequences, options, sweepers)
                if isinstance(new_result, dict):
                    result.update(new_result)

        return result

    def execute(
        self,
        sequences: list[PulseSequence],
        sweepers: Optional[list[ParallelSweepers]] = None,
        **options,
    ) -> dict[PulseId, Result]:
        """Execute pulse sequences.

        If any sweeper is passed, the execution is performed for the different values
        of sweeped parameters.

        Returns readout results acquired by after execution.

        Example:
            .. testcode::

                import numpy as np
                from qibolab import Parameter, PulseSequence, Sweeper
                from qibolab.instruments.dummy import create_dummy


                platform = create_dummy()
                qubit = platform.qubits[0]
                natives = platform.natives.single_qubit[0]
                sequence = natives.MZ.create_sequence()
                parameter_range = np.random.randint(10, size=10)
                sweeper = [
                    Sweeper(
                        parameter=Parameter.frequency,
                        values=parameter_range,
                        channels=[qubit.probe],
                    )
                ]
                platform.execute([sequence], [sweeper])
        """
        if sweepers is None:
            sweepers = []
        if not _unique_acquisitions(sequences):
            raise ValueError(
                "The acquisitions' identifiers have to be unique across all sequences."
            )

        options = self.settings.fill(ExecutionParameters(**options))

        time = options.estimate_duration(sequences, sweepers)
        log.info(f"Minimal execution time: {time:.3f} s")

        configs = self.parameters.configs.copy()
        update_configs(configs, options.updates)

        # for components that represent aux external instruments (e.g. lo) to the main
        # control instrument set the config directly
        for name, cfg in configs.items():
            if name in self.instruments:
                self.instruments[name].setup(**cfg.model_dump(exclude={"kind"}))

        results = {}
        # pylint: disable=unsubscriptable-object
        bounds = self.parameters.configs[self._controller.bounds]
        assert isinstance(bounds, Bounds)
        for b in batch(sequences, bounds):
            results |= self._execute(b, options, configs, sweepers)

        return results

    @classmethod
    def load(
        cls,
        path: Path,
        instruments: InstrumentMap,
        qubits: QubitMap,
        couplers: Optional[QubitMap] = None,
        name: Optional[str] = None,
    ) -> "Platform":
        """Dump platform."""
        parameters = Parameters.model_validate_json((path / PARAMETERS).read_text())
        return cls(
            name=name if name is not None else path.name,
            parameters=parameters,
            instruments=instruments,
            qubits=qubits,
            couplers=couplers if couplers is not None else {},
        )

    def dump(self, path: Path):
        """Dump platform."""
        (path / PARAMETERS).write_text(self.parameters.model_dump_json(indent=4))

    def _element(self, qubit: QubitId, coupler=False) -> tuple[QubitId, Qubit]:
        elements = self.qubits if not coupler else self.couplers
        try:
            return qubit, elements[qubit]
        except KeyError:
            assert isinstance(qubit, int)
            return list(self.qubits.items())[qubit]

    def qubit(self, qubit: QubitId) -> tuple[QubitId, Qubit]:
        """Retrieve physical qubit name and object.

        Temporary fix for the compiler to work for platforms where the
        qubits are not named as 0, 1, 2, ...
        """
        return self._element(qubit)

    def coupler(self, coupler: QubitId) -> tuple[QubitId, Qubit]:
        """Retrieve physical coupler name and object.

        Temporary fix for the compiler to work for platforms where the
        couplers are not named as 0, 1, 2, ...
        """
        return self._element(coupler, coupler=True)



"""Utils for pulse handling."""

from collections import defaultdict
from collections.abc import Iterable
from typing import Union

from keysight import qcs

from qibolab._core.pulses import Drag, Envelope, Gaussian, PulseId, Rectangular
from qibolab._core.pulses.pulse import PulseLike

NS_TO_S = 1e-9


def generate_qcs_envelope(shape: Envelope) -> qcs.Envelope:
    """Converts a Qibolab pulse envelope to a QCS Envelope object."""
    if isinstance(shape, Rectangular):
        return qcs.ConstantEnvelope()

    elif isinstance(shape, (Gaussian, Drag)):
        return qcs.GaussianEnvelope(shape.rel_sigma)

    else:
        # TODO: Rework this code to support other Qibolab pulse envelopes
        # raw_envelope = shape.i(num_samples) + 1j * shape.q(num_samples)
        # return qcs.ArbitraryEnvelope(
        #    times=np.linspace(0, 1, num_samples), amplitudes=raw_envelope
        # )
        raise Exception("Envelope not supported")


def process_acquisition_channel_pulses(
    program: qcs.Program,
    pulses: Iterable[PulseLike],
    frequency: Union[float, qcs.Scalar],
    virtual_channel: qcs.Channels,
    probe_virtual_channel: qcs.Channels,
    sweeper_pulse_map: defaultdict[PulseId, dict[str, qcs.Scalar]],
    classifier: qcs.Classifier = None,
):
    """Processes Qibolab pulses on the acquisition channel into QCS hardware
    instructions and adds it to the current program.

    Arguments:
        program (qcs.Program): Program object for the current sequence.
        pulses (Iterable[PulseLike]): Array of pulse objects to be processed.
        frequency (Union[float, qcs.Scalar]): Frequency of the channel.
        virtual_channel (qcs.Channels): QCS virtual digitizer channel.
        probe_virtual_channel (qcs.Channels): QCS virtual AWG channel connected to the digitzer.
        sweeper_pulse_map (defaultdict[PulseId, dict[str, qcs.Scalar]]): Map of pulse ID to map of parameter
        to be swept and corresponding QCS variable.
    """

    for pulse in pulses:
        sweep_param_map = sweeper_pulse_map.get(pulse.id, {})

        if pulse.kind == "delay":
            qcs_pulse = qcs.Delay(
                sweep_param_map.get("duration", pulse.duration * NS_TO_S)
            )
            program.add_waveform(qcs_pulse, virtual_channel)
            program.add_waveform(qcs_pulse, probe_virtual_channel)

        elif pulse.kind == "acquisition":
            duration = sweep_param_map.get("duration", pulse.duration * NS_TO_S)
            program.add_acquisition(duration, virtual_channel)

        elif pulse.kind == "readout":
            sweep_param_map = sweeper_pulse_map.get(pulse.probe.id, {})
            qcs_pulse = qcs.RFWaveform(
                duration=sweep_param_map.get(
                    "duration", pulse.probe.duration * NS_TO_S
                ),
                envelope=generate_qcs_envelope(pulse.probe.envelope),
                amplitude=sweep_param_map.get("amplitude", pulse.probe.amplitude),
                rf_frequency=frequency,
                instantaneous_phase=sweep_param_map.get(
                    "relative_phase", pulse.probe.relative_phase
                ),
            )
            integration_filter = qcs.IntegrationFilter(qcs_pulse)
            program.add_waveform(qcs_pulse, probe_virtual_channel)
            program.add_acquisition(integration_filter, virtual_channel, classifier)


def process_iq_channel_pulses(
    program: qcs.Program,
    pulses: Iterable[PulseLike],
    frequency: Union[float, qcs.Scalar],
    virtual_channel: qcs.Channels,
    sweeper_pulse_map: defaultdict[PulseId, dict[str, qcs.Scalar]],
):
    """Processes Qibolab pulses on the IQ channel into QCS hardware
    instructions and adds it to the current program.

    Arguments:
        program (qcs.Program): Program object for the current sequence.
        pulses (Iterable[PulseLike]): Array of pulse objects to be processed.
        frequency (Union[float, qcs.Scalar]): Frequency of the channel.
        virtual_channel (qcs.Channels): QCS virtual RF AWG channel.
        sweeper_pulse_map (defaultdict[PulseId, dict[str, qcs.Scalar]]): Map of pulse ID to map of parameter
        to be swept and corresponding QCS variable.
    """
    qcs_pulses = []
    for pulse in pulses:
        sweep_param_map = sweeper_pulse_map.get(pulse.id, {})

        if pulse.kind == "delay":
            qcs_pulse = qcs.Delay(
                sweep_param_map.get("duration", pulse.duration * NS_TO_S)
            )
        elif pulse.kind == "virtualz":
            qcs_pulse = qcs.PhaseIncrement(
                phase=sweep_param_map.get("relative_phase", pulse.phase)
            )
        elif pulse.kind == "pulse":
            qcs_pulse = qcs.RFWaveform(
                duration=sweep_param_map.get("duration", pulse.duration * NS_TO_S),
                envelope=generate_qcs_envelope(pulse.envelope),
                amplitude=sweep_param_map.get("amplitude", pulse.amplitude),
                rf_frequency=frequency,
                instantaneous_phase=sweep_param_map.get(
                    "relative_phase", pulse.relative_phase
                ),
            )
            if pulse.envelope.kind == "drag":
                qcs_pulse = qcs_pulse.drag(coeff=pulse.envelope.beta)
        else:
            raise ValueError("Unrecognized pulse type", pulse.kind)

        qcs_pulses.append(qcs_pulse)

    program.add_waveform(qcs_pulses, virtual_channel)


def process_dc_channel_pulses(
    program: qcs.Program,
    pulses: Iterable[PulseLike],
    virtual_channel: qcs.Channels,
    sweeper_pulse_map: defaultdict[PulseId, dict[str, qcs.Scalar]],
):
    """Processes Qibolab pulses on the DC channel into QCS hardware
    instructions and adds it to the current program.

    Arguments:
        program (qcs.Program): Program object for the current sequence.
        pulses (Iterable[PulseLike]): Array of pulse objects to be processed.
        virtual_channel (qcs.Channels): QCS virtual baseband AWG channel.
        sweeper_pulse_map (defaultdict[PulseId, dict[str, qcs.Scalar]]): Map of pulse ID to map of parameter
        to be swept and corresponding QCS variable.
    """
    qcs_pulses = []
    for pulse in pulses:
        sweep_param_map = sweeper_pulse_map.get(pulse.id, {})
        if pulse.kind == "delay":
            qcs_pulse = qcs.Delay(
                sweep_param_map.get("duration", pulse.duration * NS_TO_S)
            )
        elif pulse.kind == "pulse":
            qcs_pulse = qcs.DCWaveform(
                duration=sweep_param_map.get("duration", pulse.duration * NS_TO_S),
                envelope=generate_qcs_envelope(pulse.envelope),
                amplitude=sweep_param_map.get("amplitude", pulse.amplitude),
            )
        else:
            raise ValueError("Unrecognized pulse type", pulse.kind)
        qcs_pulses.append(qcs_pulse)

    program.add_waveform(qcs_pulses, virtual_channel)



from collections.abc import Iterable
from typing import Annotated, Optional

import numpy as np
from pydantic import PlainSerializer, PlainValidator

from qibolab._core.execution_parameters import ExecutionParameters
from qibolab._core.identifier import ChannelId
from qibolab._core.pulses import Align, Pulse, PulseLike, Readout
from qibolab._core.sequence import PulseSequence
from qibolab._core.serialize import Model
from qibolab._core.sweeper import (
    ParallelSweepers,
    Parameter,
    swept_channels,
    swept_pulses,
)

from ..q1asm import Program, parse
from .acquisition import Acquisitions, MeasureId, Weight, Weights, acquisitions
from .program import program
from .waveforms import Waveforms, waveform_indices, waveforms

__all__ = ["Q1Sequence"]


def _weight_len(w: Optional[Weight]) -> Optional[int]:
    return len(w.data) if w is not None else None


def _apply_sampling_rate(
    sequence: Iterable[PulseLike],
    sweepers: list[ParallelSweepers],
    sampling_rate: float,
) -> tuple[list[PulseLike], list[ParallelSweepers]]:
    sequence_ = [
        p.model_copy(
            update={"duration": p.duration * sampling_rate}
            if not isinstance(p, Align)
            # TODO: replace with `hasattr(p, "duration")`
            # but it's not passing the type checker
            # https://github.com/microsoft/pylance-release/issues/2237
            else {}
        )
        for p in sequence
    ]
    sweepers_ = [
        [
            (s * sampling_rate) if s.parameter is Parameter.duration else s
            for s in parsweep
        ]
        for parsweep in sweepers
    ]
    return (sequence_, sweepers_)


class Q1Sequence(Model):
    waveforms: Waveforms
    weights: Weights
    acquisitions: Acquisitions
    program: Annotated[
        Program,
        PlainSerializer(lambda p: p.asm()),
        PlainValidator(lambda p: p if isinstance(p, Program) else parse(p)),
    ]

    @classmethod
    def from_pulses(
        cls,
        sequence: list[PulseLike],
        sweepers: list[ParallelSweepers],
        options: ExecutionParameters,
        sampling_rate: float,
        channel: set[ChannelId],
        time_of_flight: Optional[float],
        duration: float,
    ) -> "Q1Sequence":
        padding = (
            duration
            - sum(p.duration if not isinstance(p, Align) else 0 for p in sequence)
        ) * sampling_rate
        waveforms_ = waveforms(
            sequence,
            sampling_rate,
            amplitude_swept={
                p.id for p in swept_pulses(sweepers, {Parameter.amplitude})
            },
            duration_swept={
                k: v
                for k, v in swept_pulses(sweepers, {Parameter.duration}).items()
                if isinstance(k, (Pulse, Readout)) and k in sequence
            },
        )
        sequence, sweepers = _apply_sampling_rate(sequence, sweepers, sampling_rate)
        acquisitions_ = acquisitions(
            sequence, np.prod(options.bins(sweepers), dtype=int)
        )
        return cls(
            waveforms={k: w.waveform for k, w in waveforms_.items()},
            weights={},
            acquisitions={k: a.acquisition for k, a in acquisitions_.items()},
            program=program(
                sequence,
                waveform_indices(waveforms_),
                acquisitions_,
                options,
                sweepers,
                channel,
                time_of_flight,
                int(padding),
            ),
        )

    @classmethod
    def empty(cls) -> "Q1Sequence":
        return cls(
            waveforms={}, weights={}, acquisitions={}, program=Program(elements=[])
        )

    @property
    def is_empty(self) -> bool:
        return len(self.program.elements) == 0

    @property
    def integration_lengths(self) -> dict[MeasureId, Optional[int]]:
        """Determine the integration lengths fixed by weights.

        Returns ``None`` for those acquisitions which are non-weighted, since the length
        is determined by a runtime configuration.

        For those, cf.
        https://docs.qblox.com/en/main/api_reference/sequencer.html#Sequencer.integration_length_acq
        """

        return {acq: _weight_len(self.weights.get(acq)) for acq in self.acquisitions}


def _effective_channels(ch: ChannelId, seq: Iterable[PulseLike]) -> set[ChannelId]:
    """Identify effective channels related to a subsequence.

    The channel is the declared one, unless in presence of :class:`Readout` operations.
    In which case the assumed *acquisition* channel is supplemented with a *probe* one.
    """
    return (
        {ch}
        if not any(isinstance(e, Readout) for e in seq)
        else {ch, f"{'/'.join(ch.split('/')[:-1])}/probe"}
    )


def compile(
    sequence: PulseSequence,
    sweepers: list[ParallelSweepers],
    options: ExecutionParameters,
    sampling_rate: float,
    time_of_flights: dict[ChannelId, float],
) -> dict[ChannelId, Q1Sequence]:
    duration = sequence.duration
    sweeper_channels = {ch: [] for ch in swept_channels(sweepers)}
    return {
        ch: Q1Sequence.from_pulses(
            seq,
            sweepers,
            options,
            sampling_rate,
            _effective_channels(ch, seq),
            time_of_flights.get(ch),
            duration,
        )
        for ch, seq in (sweeper_channels | sequence.by_channel).items()
    }



from .acquisition import (
    Acquisitions as Acquisitions,
)
from .acquisition import (
    create_acquisition as create_acquisition,
)
from .arguments import ExecutionArguments as ExecutionArguments
from .instructions import program as program



from functools import reduce
from operator import or_
from typing import Optional, TypedDict
from uuid import UUID

import numpy as np
from qblox_instruments.qcodes_drivers.module import Module

from qibolab._core.execution_parameters import AcquisitionType
from qibolab._core.identifier import ChannelId, Result
from qibolab._core.pulses.pulse import PulseId

from .identifiers import SequencerId, SequencerMap, SlotId
from .sequence import Q1Sequence, acquisition

__all__ = []


def _fill_empty_lenghts(
    weighted: dict[acquisition.MeasureId, Optional[int]],
    defaults: dict[tuple[SlotId, SequencerId], int],
    locations: dict[acquisition.MeasureId, tuple[SlotId, SequencerId]],
) -> dict[acquisition.MeasureId, int]:
    return {
        k: v if v is not None else defaults[locations[k]] for k, v in weighted.items()
    }


def integration_lenghts(
    sequences: dict[ChannelId, Q1Sequence],
    sequencers: SequencerMap,
    modules: dict[SlotId, Module],
) -> dict[acquisition.MeasureId, int]:
    channels_to_sequencer = {
        ch: (mod, seq) for mod, chs in sequencers.items() for ch, seq in chs.items()
    }
    return _fill_empty_lenghts(
        reduce(or_, (seq.integration_lengths for seq in sequences.values())),
        {
            (mod_id, i): seq.integration_length_acq()
            for mod_id, mod in modules.items()
            for i, seq in enumerate(mod.sequencers)
            if hasattr(seq, "integration_length_acq")
        },
        {
            acq: channels_to_sequencer[ch]
            for ch, seq in sequences.items()
            for acq in seq.acquisitions
        },
    )


IndividualScope = TypedDict(
    "IndividualScope",
    {"data": list[float], "out-of-range": list[bool], "avg_cnt": list[bool]},
)


class ScopeData(TypedDict):
    path0: IndividualScope
    path1: IndividualScope


class Integration(TypedDict):
    path0: list[float]
    path1: list[float]


Thresholded = list[int]


class BinData(TypedDict):
    integration: Integration
    threshold: Thresholded
    valid: list
    avg_cnt: list[int]


class Data(TypedDict):
    scope: ScopeData
    bins: BinData


class IndexedData(TypedDict):
    index: int
    acquisition: Data


AcquiredData = dict[str, IndexedData]


def _integration(data: Integration, length: int) -> Result:
    res = np.array([data["path0"], data["path1"]])
    return np.moveaxis(res, 0, -1) / length


def _scope(data: ScopeData) -> Result:
    res = np.array([data["path0"], data["path1"]])
    return np.moveaxis(res, 0, -1)


def _classification(data: Thresholded) -> Result:
    return np.array(data, dtype=int)


def extract(
    acquisitions: dict[ChannelId, AcquiredData],
    lengths: dict[acquisition.MeasureId, int],
    acquisition: AcquisitionType,
    shape: tuple[int, ...],
) -> dict[PulseId, Result]:
    # TODO: check if the `lengths` info coincide with the
    # idata["acquisition"]["bins"]["avg_cnt"]
    return {
        UUID(acq): (
            _integration(
                idata["acquisition"]["bins"]["integration"], lengths[UUID(acq)]
            )
            if acquisition is AcquisitionType.INTEGRATION
            else _classification(idata["acquisition"]["bins"]["threshold"])
            if acquisition is AcquisitionType.DISCRIMINATION
            else _scope(idata["acquisition"]["scope"])
            if acquisition is AcquisitionType.RAW
            else np.array([])
        ).reshape(shape)
        for data in acquisitions.values()
        for acq, idata in data.items()
    }



from collections.abc import Iterable

import numpy as np
from numpy.typing import NDArray

from ...execution_parameters import (
    AcquisitionType,
    AveragingMode,
    ExecutionParameters,
)
from ...identifier import ChannelId, QubitId, Result
from ...pulses import Acquisition, PulseId, Readout
from ...sequence import PulseSequence
from .engine import Operator
from .hamiltonians import HamiltonianConfig


def ndchoice(probabilities: NDArray, samples: int) -> NDArray:
    """Sample elements with n-dimensional probabilities.

    This is the n-dimensional version of :func:`np.random.choice`, which instead of
    vectorizing over the picked elements, it assumes them to be just a plain integer
    range (which in turn could be used to index a suitable array, if relevant), while it
    allows the probabilities to be higher dimensional.

    The ``probabilities`` argument specifies the set of probabilities, which are
    intended to be normalized arrays over the innermost dimension. Such that the whole
    array describe a set of ``probabilities.shape[:-1]`` discrete distributions.

    `samples` is instead the number of samples to extract from each distribution.

    .. seealso::

        Generalized from https://stackoverflow.com/a/47722393, which presents the
        two-dimensional version.
    """
    return (
        probabilities.cumsum(-1).reshape(*probabilities.shape, -1)
        > np.random.rand(*probabilities.shape[:-1], 1, samples)
    ).argmax(-2)


def shots(probabilities: NDArray, nshots: int) -> NDArray:
    """Extract shots from state |0> ... |n> probabilities.

    This function just wraps :func:`ndchoice`, taking care of creating the n-D array of
    binomial distributions, and extracting shots as the outermost dimension.
    """
    shots = ndchoice(probabilities, nshots)
    # move shots from innermost to outermost dimension
    return np.moveaxis(shots, -1, 0)


def calculate_probabilities_from_density_matrix(
    states: NDArray, subsystems: Iterable[QubitId], nsubsystems: int, d: int
) -> NDArray:
    """Compute probabilities from density matrix."""
    states_ = np.reshape(states, states.shape[:-2] + 2 * nsubsystems * (d,))
    marginal = np.einsum(
        states_,
        # TODO: the `np.array()` wrapping call is only needed because of NumPy's type
        # annotation - in practice, it also works without
        np.array([...] + list(range(nsubsystems)) * 2),
        np.array([...] + list(subsystems)),
    )
    return np.abs(marginal).reshape((*states.shape[:-2], -1))


def acquisitions(sequence: PulseSequence) -> dict[PulseId, float]:
    """Compute acquisitions' times."""
    acq = {}
    for ch in sequence.channels:
        time = 0
        for ev in sequence.channel(ch):
            if isinstance(ev, (Acquisition, Readout)):
                acq[ev.id] = time
            time += ev.duration
    return acq


def index(ch: ChannelId, hconfig: HamiltonianConfig) -> int:
    """Returns Hilbert space index from channel id."""
    if "coupler" in ch:
        target = int(ch.split("coupler_")[1].split("/")[0])
    else:
        target = int(ch.split("/")[0])
    return hconfig.hilbert_space_index(target)


def select_acquisitions(
    states: list[Operator], acquisitions: Iterable[float], times: NDArray
) -> NDArray:
    """Select density matrices from states.

    First, retrieve acquisitions, and locate them in the tlist, to
    isolate the expectations related to measurements.

    The return type should be rank-3 array, where the last two are the density
    matrices dimensions, while the first one should correspond to the acquisitions.
    """
    acq = np.array(list(acquisitions))
    samples = np.minimum(np.searchsorted(times, acq), times.size - 1)
    return np.stack([states[n].full() for n in samples])


def results(
    states: NDArray,
    sequence: PulseSequence,
    hamiltonian: HamiltonianConfig,
    options: ExecutionParameters,
) -> dict[int, Result]:
    """Collect results for a single pulse sequence.

    The dictionary returned is already compliant with the expected
    result for the execution of this single sequence, thus suitable
    to be returned as is.
    """
    probabilities = calculate_probabilities_from_density_matrix(
        states,
        range(hamiltonian.nqubits),
        hamiltonian.nqubits,
        hamiltonian.transmon_levels,
    )

    assert options.nshots is not None
    sampled = shots(np.moveaxis(probabilities, -2, 0), options.nshots)
    # move measurements dimension to the front, getting ready for extraction
    measurements = np.moveaxis(sampled, 1, 0)

    results = {}
    # introduce cached measurements to avoid losing correlations
    cache_measurements = {}
    for (ro_id, sample), meas in zip(acquisitions(sequence).items(), measurements):
        i = index(sequence.pulse_channels(ro_id)[0], hamiltonian)
        cache_measurements.setdefault(sample, meas)
        res = np.stack(np.unravel_index(cache_measurements[sample], hamiltonian.dims))[
            i
        ]

        if options.acquisition_type is AcquisitionType.INTEGRATION:
            res = np.stack((res, np.zeros_like(res)), axis=-1)
            res = np.random.normal(res, scale=0.001)

        if options.averaging_mode == AveragingMode.CYCLIC:
            res = np.mean(res, axis=0)

        results[ro_id] = res

    return results



from typing import Any

from . import _core, _version, instruments
from ._core import *
from ._version import *

__all__ = []
__all__ += _core.__all__
__all__ += _version.__all__
__all__ += ["instruments"]


def __getattr__(name: str) -> Any:
    """Redirect meta-backend lookup for Qibo decoupling.

    Since it should be possible to use Qibolab without Qibo installed, especially to
    increase its compatibility with other packages (possibly incompatible with Qibo
    itself), the backend is made optional.

    However, the meta-backend mechanism looks for a variable identifier ``MetaBackend``
    in the top-most scope of the package.
    In order to preserve compatibility with this protocol, the variable is dynamically
    imported, if needed.

    .. note::

        Usage of this import is discouraged, and the import from
        :class:`qibolab.backend.MetaBackend` should be favored, for direct usage.

        This dynamic import may be phased out at a later stage. If the meta-backend
        protocol will allow for some different redirection.
    """
    if name == "MetaBackend":
        from .backend import MetaBackend

        return MetaBackend

    return globals()[name]



import shutil
import tempfile
import warnings
from collections import defaultdict
from dataclasses import asdict, dataclass
from os import PathLike
from pathlib import Path
from typing import Optional, Union

from pydantic import Field
from qm import QmPendingJob, QuantumMachine, QuantumMachinesManager, generate_qua_script
from qm.api.v2.qm_api_old import QmApiWithDeprecations
from qm.simulate.credentials import create_credentials

from qibolab._core.components import (
    AcquisitionChannel,
    Config,
    DcChannel,
    IqChannel,
    IqConfig,
    OscillatorConfig,
)
from qibolab._core.execution_parameters import ExecutionParameters
from qibolab._core.identifier import ChannelId
from qibolab._core.instruments.abstract import Controller
from qibolab._core.pulses import Align, Delay, Pulse, Readout
from qibolab._core.sequence import PulseSequence
from qibolab._core.serialize import Model
from qibolab._core.sweeper import ParallelSweepers, Parameter, Sweeper
from qibolab._core.unrolling import unroll_sequences

from .components import MwFemOscillatorConfig, OpxOutputConfig, QmAcquisitionConfig
from .config import Configuration, ControllerId, ModuleTypes
from .program import Acquisitions, ExecutionArguments, create_acquisition, program
from .program.sweepers import find_lo_frequencies, sweeper_amplitude

CALIBRATION_DB = "calibration_db.json"
"""Name of the file where the mixer calibration is stored."""

__all__ = ["QmController", "Octave"]

MAX_VOLTAGE = 0.5
"""Maximum output of Quantum Machines OPX+ in Volts."""
MAX_VOLTAGE_AMPLIFIED = 2.5
"""Maximum output of Quantum Machines OPX1000 FEMs in amplified mode in
Volts."""
SAMPLING_RATE = 1
"""Sampling rate of Quantum Machines OPX+ in GSps."""


def channel_max_voltage(config: Config):
    """Find maximum voltage of a channel from the associated ``config``."""
    if hasattr(config, "output_mode") and config.output_mode == "amplified":
        return MAX_VOLTAGE_AMPLIFIED
    return MAX_VOLTAGE


def channel_sampling_rate(config: Config):
    if hasattr(config, "sampling_rate"):
        return int(config.sampling_rate / 1e9)
    return SAMPLING_RATE


@dataclass(frozen=True)
class Octave:
    """User-facing object for defining Octave configuration."""

    name: str
    """Name of the device."""
    port: int
    """Network port of the Octave in the cluster configuration."""
    connectivity: ControllerId
    """OPXplus that acts as the waveform generator for the Octave."""


def fetch_results(handles, acquisitions):
    """Fetches results from an executed experiment.

    Args:
        result: Result of the executed experiment.
        acquisition: Dictionary containing :class:`qibolab.instruments.qm.acquisition.Acquisition` objects.

    Returns:
        Dictionary with the results in the format required by the platform.
    """
    results = defaultdict(list)
    for acquisition in acquisitions:
        data = acquisition.fetch(handles)
        for serial, result in zip(acquisition.keys, data):
            results[serial].append(result)

    # collapse single element lists for back-compatibility
    return {
        key: value[0] if len(value) == 1 else value for key, value in results.items()
    }


def find_sweepers(
    sweepers: list[ParallelSweepers], parameter: Parameter
) -> list[Sweeper]:
    """Find sweepers of given parameter in order to register specific pulses.

    Duration and amplitude sweepers may require registering additional pulses
    in the QM ``config``.
    """
    return [s for ps in sweepers for s in ps if s.parameter is parameter]


class Experiment(Model):
    configs: dict[str, Config]
    sequences: list[PulseSequence]
    sweepers: list[ParallelSweepers]


class Cache(Model):
    machine: Union[QuantumMachine, QmApiWithDeprecations]
    program_id: str
    acquisitions: Acquisitions

    def run(self) -> QmPendingJob:
        return self.machine.queue.add_compiled(self.program_id)


def _update_pulse_amplitude(
    pulse: Union[Pulse, Readout], amplitude: float
) -> Union[Pulse, Readout]:
    """Update pulse amplitude.

    Needed when sweeping amplitude because the original amplitude
    may not sufficient to reach all the sweeper values.
    """
    if isinstance(pulse, Pulse):
        return pulse.model_copy(update={"amplitude": amplitude})
    probe = pulse.probe.model_copy(update={"amplitude": amplitude})
    return pulse.model_copy(update={"probe": probe})


class QmController(Controller):
    """:class:`qibolab.instruments.abstract.Controller` object for controlling
    a Quantum Machines cluster.

    Playing pulses on QM controllers requires a ``config`` dictionary and a program
    written in QUA language.
    The ``config`` file is generated using the ``dataclass`` objects defined in
    :mod:`qibolab.instruments.qm.config`.
    The QUA program is generated using the methods in :mod:`qibolab.instruments.qm.program`.
    Controllers, elements and pulses are added in the ``config`` after a pulse sequence is given,
    so that only elements related to the participating channels are registered.
    """

    address: str
    """IP address and port for connecting to the OPX instruments.

    Has the form XXX.XXX.XXX.XXX:XXX.
    """

    octaves: dict[str, Octave] = Field(default_factory=dict)
    """Dictionary containing the Octaves used."""
    fems: dict[str, ModuleTypes] = Field(
        default_factory=lambda: defaultdict(lambda: "opx1")
    )
    """Dictionary containing the FEM types (for OPX1000 clusters).

    Defaults to 'opx1' type to maintain original behavior for OPX+ clusters
    where ``fems`` are not given.
    """

    keep_dc_offsets_on: bool = True
    """Keep DC offsets on after disconnecting from Quantum Machines."""
    cluster_name: Optional[str] = None
    """Name of the Quantum Machines clusters to use.

    Needs to be specified only when more than one clusters are connected to
    the same router. See
    https://docs.quantum-machines.co/latest/docs/Hardware/network_and_router/#accessing-the-cluster
    for more details.
    """
    bounds: str = "qm/bounds"
    """Maximum bounds used for batching in sequence unrolling."""
    calibration_path: Optional[PathLike] = None
    """Path to the JSON file that contains the mixer calibration."""
    write_calibration: bool = False
    """Require writing permissions on calibration DB."""
    _calibration_path: Optional[PathLike] = None
    """The calibration path for internal use.

    Cf. :attr:`calibration_path` for its role. This might be set to a different one
    internally to avoid writing attempts over a file for which the user has only read
    access (because TinyDB, through QUA, is often attempting to open it in append mode).
    """
    script_file_name: Optional[str] = None
    """Name of the file that the QUA program will dumped in that after every
    execution.

    If ``None`` the program will not be dumped.
    """

    manager: Optional[QuantumMachinesManager] = None
    """Manager object used for controlling the Quantum Machines cluster."""

    config: Configuration = Field(default_factory=Configuration)
    """Configuration dictionary required for pulse execution on the OPXs."""
    experiment: Optional[Experiment] = None
    cache: Optional[Cache] = None

    simulation_duration: Optional[int] = None
    """Duration for the simulation in ns.

    If given the simulator will be used instead of actual hardware
    execution.
    """
    cloud: bool = False
    """If ``True`` the QM cloud simulator is used which does not require access
    to physical instruments.

    This assumes that a proper cloud address has been given.
    If ``False`` and ``simulation_duration`` was given, then the built-in simulator
    of the instruments is used. This requires connection to instruments.
    Default is ``False``.
    """

    def model_post_init(self, __context):
        if self.simulation_duration is not None:
            raise NotImplementedError(
                "Simulation is no longer supported by the QM driver."
            )

    @property
    def sampling_rate(self):
        """Sampling rate of Quantum Machines instruments."""
        return SAMPLING_RATE

    def _temporary_calibration(self):
        if self.calibration_path is not None:
            if self.write_calibration:
                self._calibration_path = self.calibration_path
            else:
                self._calibration_path = tempfile.mkdtemp()
                shutil.copy(
                    Path(self.calibration_path) / CALIBRATION_DB,
                    Path(self._calibration_path) / CALIBRATION_DB,
                )

    def _reset_temporary_calibration(self):
        if self._calibration_path != self.calibration_path:
            assert self._calibration_path is not None
            shutil.rmtree(self._calibration_path)
            self._calibration_path = None

    def connect(self):
        """Connect to the Quantum Machines manager."""
        host, port = self.address.split(":")
        self._temporary_calibration()
        credentials = None
        if self.cloud:
            credentials = create_credentials()
        self.manager = QuantumMachinesManager(
            host=host,
            port=int(port),
            credentials=credentials,
            cluster_name=self.cluster_name,
            octave_calibration_db_path=self._calibration_path,
        )

    def disconnect(self):
        """Disconnect from QM manager."""
        self._reset_temporary_calibration()
        if self.manager is not None:
            if not self.keep_dc_offsets_on:
                self.manager.close_all_quantum_machines()
            self.manager = None

    def configure_device(self, device: str):
        """Add device in the ``config``."""
        if "oct" in device:
            self.config.add_octave(device, self.octaves[device].connectivity, self.fems)
        else:
            self.config.add_controller(device, self.fems)

    def configure_channel(
        self, channel: ChannelId, configs: dict[str, Config]
    ) -> Optional[ChannelId]:
        """Add element (QM version of channel) in the config.

        When an ``AcquisitionChannel`` is registered it returns the corresponding probe
        channel in order to build an (acquisition, probe) map.
        """
        config = configs[channel]
        ch = self.channels[channel]
        self.configure_device(ch.device)

        if isinstance(ch, DcChannel):
            assert isinstance(config, OpxOutputConfig)
            self.config.configure_dc_line(channel, ch, config)

        elif isinstance(ch, IqChannel):
            assert isinstance(config, IqConfig)
            assert ch.lo is not None
            lo_config = configs[ch.lo]
            assert isinstance(lo_config, OscillatorConfig)
            if isinstance(lo_config, MwFemOscillatorConfig):
                self.config.configure_mw_fem_line(ch, config, lo_config, channel)
            else:
                self.config.configure_iq_line(ch, config, lo_config, channel)

        elif isinstance(ch, AcquisitionChannel):
            assert ch.probe is not None
            assert isinstance(config, QmAcquisitionConfig)
            probe = self.channels[ch.probe]
            probe_config = configs[ch.probe]
            assert isinstance(probe, IqChannel)
            assert isinstance(probe_config, IqConfig)
            assert probe.lo is not None
            lo_config = configs[probe.lo]
            assert isinstance(lo_config, OscillatorConfig)
            if isinstance(lo_config, MwFemOscillatorConfig):
                self.config.configure_mw_fem_acquire_line(
                    ch,
                    probe,
                    config,
                    probe_config,
                    lo_config,
                    channel,
                )
            else:
                self.config.configure_acquire_line(
                    ch,
                    probe,
                    config,
                    probe_config,
                    lo_config,
                    channel,
                )
            return ch.probe

        else:
            raise TypeError(f"Unknown channel type: {type(ch)}.")

        return None

    def configure_channels(
        self, configs: dict[str, Config], channels: set[ChannelId]
    ) -> dict[ChannelId, ChannelId]:
        """Register channels in the sequence in the QM ``config``.

        Builds a map from probe channels to the corresponding ``AcquisitionChannel``.
        This is useful when sweeping frequency of probe channels, as these are
        registered as acquire elements in the QM config.
        """
        probe_map = {}
        for id in channels:
            probe = self.configure_channel(id, configs)
            if probe is not None:
                probe_map[probe] = id
        return probe_map

    def register_pulse(
        self, channel: ChannelId, config: Config, pulse: Union[Pulse, Readout]
    ) -> str:
        """Add pulse in the QM ``config``.

        And return corresponding operation.
        """
        ch = self.channels[channel]
        max_voltage = channel_max_voltage(config)
        sampling_rate = channel_sampling_rate(config)
        if isinstance(ch, DcChannel):
            assert isinstance(pulse, Pulse)
            return self.config.register_dc_pulse(
                channel, pulse, sampling_rate, max_voltage
            )
        if isinstance(ch, IqChannel):
            assert isinstance(pulse, Pulse)
            return self.config.register_iq_pulse(
                channel, pulse, sampling_rate, max_voltage
            )
        assert isinstance(pulse, Readout)
        return self.config.register_acquisition_pulse(
            channel, pulse, sampling_rate, max_voltage
        )

    def register_pulses(self, configs: dict[str, Config], sequence: PulseSequence):
        """Adds all pulses except measurements of a given sequence in the QM
        ``config``.

        Returns:
            acquisitions (dict): Map from measurement instructions to acquisition objects.
        """
        for id, pulse in sequence:
            if isinstance(pulse, Pulse):
                self.register_pulse(id, configs[id], pulse)
            elif isinstance(pulse, Readout):
                self.register_pulse(id, configs[id], pulse)

    def register_duration_sweeper_pulses(
        self, args: ExecutionArguments, configs: dict[str, Config], sweeper: Sweeper
    ):
        """Register pulses needed to implement a duration sweep.

        For standard (non-interpolated) duration sweeps, we need to
        upload distinct waveforms for every different duration.

        When interpolation is used, we need to upload a single pulse
        with the minimum duration of the sweeper, because the QM real-
        time interpolation can only stretch and not compress arbitrary
        waveforms, as documented in
        https://docs.quantum-machines.co/latest/docs/Guides/features/?h=interpo#dynamic-pulse-duration
        """
        for pulse in sweeper.pulses:
            if isinstance(pulse, (Align, Delay)):
                continue

            ids = args.sequence.pulse_channels(pulse.id)
            sampling_rate = channel_sampling_rate(configs[ids[0]])

            if not abs(sweeper.values[1] - sweeper.values[0]).is_integer():
                assert sweeper.parameter is Parameter.duration
                assert sampling_rate > 1
            for _id in ids:
                assert channel_sampling_rate(configs[_id]) == sampling_rate

            params = args.parameters[pulse.id]

            params.sampling_rate = sampling_rate
            original_pulse = (
                pulse if params.amplitude_pulse is None else params.amplitude_pulse
            )

            if sweeper.parameter is Parameter.duration_interpolated:
                sweep_pulse = original_pulse.model_copy(
                    update={"duration": int(min(sweeper.values))}
                )
                params.interpolated_op = self.register_pulse(
                    ids[0], configs[ids[0]], sweep_pulse
                )
            else:
                assert sweeper.parameter is Parameter.duration
                for duration in sweeper.values:
                    sweep_pulse = original_pulse.model_copy(
                        update={"duration": duration}
                    )
                    sweep_op = self.register_pulse(ids[0], configs[ids[0]], sweep_pulse)
                    nsamples = int(sampling_rate * duration)
                    params.duration_ops.append((nsamples, sweep_op))

    def register_amplitude_sweeper_pulses(
        self, args: ExecutionArguments, configs: dict[str, Config], sweeper: Sweeper
    ):
        """Register pulse with different amplitude.

        Needed when sweeping amplitude because the original amplitude
        may not sufficient to reach all the sweeper values.
        """
        amplitude = sweeper_amplitude(sweeper.values)
        for pulse in sweeper.pulses:
            sweep_pulse = _update_pulse_amplitude(pulse, amplitude)
            ids = args.sequence.pulse_channels(pulse.id)
            params = args.parameters[pulse.id]
            params.amplitude_pulse = sweep_pulse
            params.amplitude_op = self.register_pulse(
                ids[0], configs[ids[0]], sweep_pulse
            )

    def register_acquisitions(
        self,
        configs: dict[str, Config],
        sequence: PulseSequence,
        options: ExecutionParameters,
    ) -> Acquisitions:
        """Add all measurements of a given sequence in the QM ``config``.

        Returns:
            acquisitions (dict): Map from measurement instructions to acquisition objects.
        """
        acquisitions = {}
        for channel_id, readout in sequence:
            if not isinstance(readout, Readout):
                continue

            probe_id = self.channels[channel_id].probe
            max_voltage = channel_max_voltage(configs[probe_id])
            sampling_rate = channel_sampling_rate(configs[probe_id])
            op = self.config.register_acquisition_pulse(
                channel_id, readout, sampling_rate, max_voltage
            )

            acq_config = configs[channel_id]
            assert isinstance(acq_config, QmAcquisitionConfig)
            self.config.register_integration_weights(
                channel_id, readout.duration, acq_config.kernel
            )
            if (op, channel_id) in acquisitions:
                acquisition = acquisitions[(op, channel_id)]
            else:
                acquisition = acquisitions[(op, channel_id)] = create_acquisition(
                    op, channel_id, options, acq_config.threshold, acq_config.iq_angle
                )
            acquisition.keys.append(readout.acquisition.id)

        return acquisitions

    def preprocess_sweeps(
        self,
        sweepers: list[ParallelSweepers],
        configs: dict[str, Config],
        args: ExecutionArguments,
        probe_map: dict[ChannelId, ChannelId],
    ):
        """Preprocessing and checks needed before executing some sweeps.

        Amplitude and duration sweeps require registering additional pulses in the QM
        ``config``.
        """
        for sweeper in find_sweepers(sweepers, Parameter.frequency):
            channels = [(id, self.channels[id]) for id in sweeper.channels]
            find_lo_frequencies(args, channels, configs, sweeper.values)
            for id in sweeper.channels:
                args.parameters[id].element = probe_map.get(id, id)
        for sweeper in find_sweepers(sweepers, Parameter.offset):
            for id in sweeper.channels:
                args.parameters[id].element = id
                args.parameters[id].max_offset = channel_max_voltage(configs[id])
        for sweeper in find_sweepers(sweepers, Parameter.amplitude):
            self.register_amplitude_sweeper_pulses(args, configs, sweeper)
        for sweeper in find_sweepers(sweepers, Parameter.duration):
            self.register_duration_sweeper_pulses(args, configs, sweeper)
        for sweeper in find_sweepers(sweepers, Parameter.duration_interpolated):
            self.register_duration_sweeper_pulses(args, configs, sweeper)

    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ):
        if len(sequences) == 0:
            return {}
        elif len(sequences) == 1:
            sequence = sequences[0]
        else:
            sequence, _ = unroll_sequences(sequences, options.relaxation_time)

        if len(sequence) == 0:
            return {}

        new_experiment = Experiment(
            configs={ch: configs[ch] for ch in configs.keys() & self.channels.keys()},
            sequences=sequences,
            sweepers=sweepers,
        )
        if new_experiment != self.experiment or self.manager is None:
            # register DC elements so that all qubits are
            # sweetspot even when they are not used
            for id, channel in self.channels.items():
                if isinstance(channel, DcChannel):
                    self.configure_channel(id, configs)

            probe_map = self.configure_channels(configs, sequence.channels)
            self.register_pulses(configs, sequence)
            acquisitions = self.register_acquisitions(configs, sequence, options)

            args = ExecutionArguments(sequence, acquisitions, options.relaxation_time)
            self.preprocess_sweeps(sweepers, configs, args, probe_map)
            qua_program = program(args, options, sweepers)

            if self.script_file_name is not None:
                script = generate_qua_script(qua_program, asdict(self.config))
                with open(self.script_file_name, "w") as file:
                    file.write(script)

            if self.manager is None:
                warnings.warn(
                    "Not connected to Quantum Machines. Returning program and config."
                )
                return {"program": qua_program, "config": asdict(self.config)}

            machine = self.manager.open_qm(asdict(self.config))
            program_id = machine.compile(qua_program)
            self.cache = Cache(
                machine=machine, program_id=program_id, acquisitions=acquisitions
            )
            self.experiment = new_experiment

        pending_job = self.cache.run()
        job = pending_job.wait_for_execution()
        handles = job.result_handles
        handles.wait_for_all_values()  # for async replace with ``handles.is_processing()``
        return fetch_results(handles, self.cache.acquisitions.values())



from dataclasses import dataclass, field
from typing import Union

from typing_extensions import TypedDict

from qibolab._core.components import Channel

__all__ = [
    "DcElement",
    "RfOctaveElement",
    "AcquireOctaveElement",
    "MwFemElement",
    "AcquireMwFemElement",
    "Element",
]


InOutType = Union[tuple[str, int], tuple[str, int, int]]
OctavePort = TypedDict("OpxPlusPort", {"port": tuple[str, int]})
Port = TypedDict("Port", {"port": InOutType})
ConnectivityType = Union[str, tuple[str, int]]


@dataclass(frozen=True)
class OutputSwitch:
    port: InOutType
    delay: int = 57
    buffer: int = 18
    """Default calibration parameters for digital pulses.

    https://docs.quantum-machines.co/1.1.7/qm-qua-sdk/docs/Guides/octave/#calibrating-the-digital-pulse

    Digital markers are used for LO triggering.
    """


def _to_port(channel: Channel) -> Port:
    """Convert a channel to the port dictionary required for the QUA config.

    The following syntax is assumed for ``channel.device``:
    * For OPX+ clusters: string with the device name (eg. 'con1')
    * For OPX1000 clusters: string of '{device_name}/{fem_number}'
    """
    if "/" not in channel.device:
        port = (channel.device, channel.port)
    else:
        con, fem = channel.device.split("/")
        port = (con, int(fem), channel.port)
    return {"port": port}


def output_switch(connectivity: ConnectivityType, port: int):
    """Create output switch section."""
    if isinstance(connectivity, tuple):
        args = connectivity + (2 * port - 1,)
    else:
        args = (connectivity, 2 * port - 1)
    return {"output_switch": OutputSwitch(args)}


@dataclass
class DcElement:
    singleInput: Port
    intermediate_frequency: int = 0
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(cls, channel: Channel):
        return cls(_to_port(channel))


DigitalInputs = TypedDict("digitalInputs", {"output_switch": OutputSwitch})


@dataclass
class RfOctaveElement:
    RF_inputs: OctavePort
    digitalInputs: DigitalInputs
    intermediate_frequency: int
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(
        cls,
        channel: Channel,
        connectivity: ConnectivityType,
        intermediate_frequency: int,
    ):
        return cls(
            _to_port(channel),
            output_switch(connectivity, channel.port),
            intermediate_frequency,
        )


@dataclass
class AcquireOctaveElement:
    RF_inputs: OctavePort
    RF_outputs: OctavePort
    digitalInputs: DigitalInputs
    intermediate_frequency: int
    time_of_flight: int = 24
    smearing: int = 0
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(
        cls,
        probe_channel: Channel,
        acquire_channel: Channel,
        connectivity: ConnectivityType,
        intermediate_frequency: int,
        time_of_flight: int,
        smearing: int,
    ):
        return cls(
            _to_port(probe_channel),
            _to_port(acquire_channel),
            output_switch(connectivity, probe_channel.port),
            intermediate_frequency,
            time_of_flight=time_of_flight,
            smearing=smearing,
        )


MwInput = TypedDict("MwInput", {"port": tuple[str, int, int], "upconverter": int})
MwOutput = TypedDict("MwOutput", {"port": tuple[str, int, int]})


def _to_mw_fem_input(channel: Channel, upconverter: int):
    return _to_port(channel) | {"upconverter": upconverter}


@dataclass
class MwFemElement:
    MWInput: MwInput
    intermediate_frequency: int
    digitalInputs: DigitalInputs = field(default_factory=dict)
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(
        cls, channel: Channel, upconverter: int, intermediate_frequency: int
    ):
        return cls(
            _to_mw_fem_input(channel, upconverter),
            intermediate_frequency=intermediate_frequency,
        )


@dataclass
class AcquireMwFemElement:
    MWInput: MwInput
    MWOutput: MwOutput
    intermediate_frequency: int
    digitalInputs: DigitalInputs = field(default_factory=dict)
    time_of_flight: int = 24
    smearing: int = 0
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(
        cls,
        probe_channel: Channel,
        upconverter: int,
        acquire_channel: Channel,
        intermediate_frequency: int,
        time_of_flight: int,
        smearing: int,
    ):
        return cls(
            _to_mw_fem_input(probe_channel, upconverter),
            _to_port(acquire_channel),
            intermediate_frequency=intermediate_frequency,
            time_of_flight=time_of_flight,
            smearing=smearing,
        )


Element = Union[
    DcElement,
    RfOctaveElement,
    AcquireOctaveElement,
    MwFemElement,
    AcquireMwFemElement,
]



"""PulseSequence class."""

from collections import UserList, defaultdict
from collections.abc import Callable, Iterable
from functools import cache
from typing import Any, Union

from pydantic import TypeAdapter
from pydantic_core import core_schema

from qibolab._core.pulses.pulse import PulseId

from .identifier import ChannelId
from .pulses import Acquisition, Align, Delay, PulseLike, Readout

__all__ = ["PulseSequence"]

_Element = tuple[ChannelId, PulseLike]
InputOps = Union[Readout, Acquisition]

_adapted_sequence = TypeAdapter(list[_Element])


def _synchronize(sequence: "PulseSequence", channels: Iterable[ChannelId]) -> None:
    """Helper for ``concatenate`` and ``align_to_delays``.

    Modifies given ``sequence`` in-place!
    """
    durations = {ch: sequence.channel_duration(ch) for ch in channels}
    max_duration = max(durations.values(), default=0.0)
    for ch, duration in durations.items():
        delay = max_duration - duration
        if delay > 0:
            sequence.append((ch, Delay(duration=delay)))


class PulseSequence(UserList[_Element]):
    """Synchronized sequence of control instructions across multiple channels.

    The sequence is a linear stream of instructions, which may be
    executed in parallel over multiple channels.

    Each instruction is composed by the pulse-like object representing
    the action, and the channel on which it should be performed.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        schema = handler(list[_Element])
        return core_schema.no_info_after_validator_function(
            cls._validate,
            schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls._serialize, info_arg=False
            ),
        )

    @classmethod
    def _validate(cls, value):
        return cls(value)

    @staticmethod
    def _serialize(value):
        return _adapted_sequence.dump_python(list(value))

    @classmethod
    def load(cls, value: list[tuple[str, PulseLike]]):
        return TypeAdapter(cls).validate_python(value)

    @property
    def duration(self) -> float:
        """Duration of the entire sequence."""
        return max((self.channel_duration(ch) for ch in self.channels), default=0.0)

    @property
    def channels(self) -> set[ChannelId]:
        """Channels involved in the sequence."""
        return {ch for (ch, _) in self}

    def channel(self, channel: ChannelId) -> Iterable[PulseLike]:
        """Isolate pulses on a given channel."""
        return (pulse for (ch, pulse) in self if ch == channel)

    def channel_duration(self, channel: ChannelId) -> float:
        """Duration of the given channel."""
        sequence = (
            self.align_to_delays()
            if any(isinstance(pulse, Align) for _, pulse in self)
            else self
        )
        return sum(pulse.duration for pulse in sequence.channel(channel))

    def pulse_channels(self, pulse_id: PulseId) -> list[ChannelId]:
        """Find channels on which a pulse with a given id plays."""
        return [channel for channel, pulse in self if pulse.id == pulse_id]

    def concatenate(self, other: Iterable[_Element]) -> None:
        """Concatenate two sequences.

        Appends ``other`` in-place such that the result is:

            - ``self``
            - necessary delays to synchronize channels
            - ``other``

        Guarantees that the all the channels in the concatenated
        sequence will start simultaneously
        """
        _synchronize(self, PulseSequence(other).channels)
        self.extend(other)

    def __ilshift__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        Alias to :meth:`concatenate`.
        """
        self.concatenate(other)
        return self

    def __lshift__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        A copy is made, and no input is altered.

        Other than that, it is based on :meth:`concatenate`.
        """
        copy = self.copy()
        copy <<= other
        return copy

    def juxtapose(self, other: Iterable[_Element]) -> None:
        """Juxtapose two sequences.

        Appends ``other`` in-place such that the result is:

            - ``self``
            - necessary delays to synchronize channels
            - ``other``

        Guarantee simultaneous start and no overlap.
        """
        _synchronize(self, PulseSequence(other).channels | self.channels)
        self.extend(other)

    def __ior__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        Alias to :meth:`concatenate`.
        """
        self.juxtapose(other)
        return self

    def __or__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        A copy is made, and no input is altered.

        Other than that, it is based on :meth:`concatenate`.
        """
        copy = self.copy()
        copy |= other
        return copy

    def align(self, channels: list[ChannelId]) -> Align:
        """Introduce align commands to the sequence."""
        align = Align()
        for channel in channels:
            self.append((channel, align))
        return align

    def align_to_delays(self) -> "PulseSequence":
        """Compile align commands to delays."""

        # keep track of ``Align`` command that were already played
        # because the same ``Align`` will appear on multiple channels
        # in the sequence
        processed_aligns = set()

        new = type(self)()
        for channel, pulse in self:
            if isinstance(pulse, Align):
                if pulse.id not in processed_aligns:
                    channels = self.pulse_channels(pulse.id)
                    _synchronize(new, channels)
                    processed_aligns.add(pulse.id)
            else:
                new.append((channel, pulse))
        return new

    def trim(self) -> "PulseSequence":
        """Drop final delays.

        The operation is not in place, and does not modify the original
        sequence.
        """
        terminated = set()
        new = []
        for ch, pulse in reversed(self):
            if ch not in terminated:
                if isinstance(pulse, Delay):
                    continue
                terminated.add(ch)
            new.append((ch, pulse))
        return type(self)(reversed(new))

    @property
    def acquisitions(self) -> list[tuple[ChannelId, InputOps]]:
        """Return list of the readout pulses in this sequence.

        .. note::

            This selects only the :class:`Acquisition` events, and not all the
            instructions directed to an acquistion channel
        """
        # pulse filter needed to exclude delays
        return [(ch, p) for ch, p in self if isinstance(p, (Acquisition, Readout))]

    @property
    def split_readouts(self) -> "PulseSequence":
        """Split readout operations in its constituents.

        This will also double the rest of the channels (mainly delays) on which the
        readouts are placed, assuming the probe channels to be absent.

        .. note::

            Since :class:`Readout` is only placed on an acquisition channel, the name of
            the associated probe channel is actually unknown.
            This function assumes the convention that the relevant channels are named
            ``.../acquisition`` and ``.../probe``.
        """

        def unwrap(pulse: PulseLike, double: bool) -> tuple[PulseLike, ...]:
            return (
                (pulse.acquisition, pulse.probe)
                if isinstance(pulse, Readout)
                else (pulse, pulse)
                if double
                else (pulse,)
            )

        @cache
        def probe(channel: ChannelId) -> ChannelId:
            return channel.split("/")[0] + "/probe"

        readouts = {ch for ch, p in self if isinstance(p, Readout)}
        return type(self)(
            [
                (ch_, p_)
                for ch, p in self
                for ch_, p_ in zip((ch, probe(ch)), unwrap(p, ch in readouts))
            ]
        )

    @property
    def by_channel(self) -> dict[ChannelId, list[PulseLike]]:
        """Separate sequence into channels dictionary."""
        seqs = defaultdict(list)
        for ch, pulse in self:
            seqs[ch].append(pulse)

        return seqs



import json
from typing import Optional, cast

import numpy as np
from pydantic import ConfigDict
from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qibolab._core.components.channels import Channel, IqChannel
from qibolab._core.components.configs import (
    AcquisitionConfig,
    Configs,
    DcConfig,
    IqConfig,
    OscillatorConfig,
)
from qibolab._core.execution_parameters import AcquisitionType
from qibolab._core.identifier import ChannelId
from qibolab._core.serialize import Model

from ..q1asm.ast_ import Acquire, Line
from ..sequence import Q1Sequence
from .port import PortAddress

__all__ = []


def _integration_length(sequence: Q1Sequence) -> Optional[int]:
    """Find integration length based on sequence waveform lengths."""
    lengths = {
        line.instruction.duration
        for line in sequence.program.elements
        if isinstance(line, Line)
        if isinstance(line.instruction, Acquire)
    }
    if len(lengths) == 0:
        return None
    if len(lengths) == 1:
        return lengths.pop()
    raise NotImplementedError(
        "Cannot acquire different lengths using the same sequencer."
    )


QCM_SWEEP_TO_OFFSET = 2.5 / np.sqrt(2)
"""Conversion factor between swept value and configuration.

There are two different ways to add an offset to the waveform played by the QCM module:

- digitally summing an offset, which could be controlled both in real-time and by
  conifgurations
- adding an offset directly to the outcoming signal


Since the QCM supplies outputs at 5 Vpp (`documented as +/-2.5 V
<https://docs.qblox.com/en/main/products/architecture/modules/qcm.html#specifications>`_),
a conversion is neeeded, because the first option will be defined in the interval (-1,
1) in the parameters (internally mapping the floats on a suitable integers range), while
the second is directly expressed in Volt.
Hence, the conversion factor of ``2.5``.

However, these two ways are not equivalent, especially because of the NCO and LO mixing
process.
Indeed, the first one is happening upstream to the mixing process, and the second
downstream.
Since we are sweeping only one of the two components of the signal (the in-phase,
I), it will result multiplied by a sine-wave, which reduces its root mean square (RMS)
power by a factor of `sqrt(2)`. Which is then accounted for in the conversion range.

https://docs.qblox.com/en/main/products/architecture/sequencers/control.html#arbitrary-waveform-generator-awg
https://docs.qblox.com/en/main/products/architecture/modules/qcm.html#block-diagram

Notice that sweeping both of the components is also viable. But even without any flux
pulse, the sum of sine and cosine with maximal amplitude will saturate the power supply,
eventually clipping the signal and reducing the power range.
"""


class SequencerConfig(Model):
    # disable freeze, to be able to construct instance with optional fields, but also
    # static validation
    model_config = ConfigDict(frozen=False)

    module: dict
    address: Optional[str] = None
    # the following attributes are automatically processed and set
    sequence: Optional[dict] = None
    sync_en: Optional[bool] = None
    offset_awg_path0: Optional[float] = None
    offset_awg_path1: Optional[float] = None
    marker_ovr_en: Optional[bool] = None
    marker_ovr_value: Optional[int] = None
    integration_length_acq: Optional[int] = None
    thresholded_acq_rotation: Optional[float] = None
    thresholded_acq_threshold: Optional[float] = None
    demod_en_acq: Optional[bool] = None
    nco_freq: Optional[int] = None
    mod_en_awg: Optional[bool] = None

    @classmethod
    def build(
        cls,
        address: PortAddress,
        sequence: Q1Sequence,
        channel_id: ChannelId,
        channels: dict[ChannelId, Channel],
        configs: Configs,
        acquisition: AcquisitionType,
        index: int,
        rf: bool,
    ) -> "SequencerConfig":
        module = {}
        config = configs[channel_id]

        # set parameters
        # offsets
        if isinstance(config, DcConfig):
            module[f"out{index}_offset"] = config.offset * QCM_SWEEP_TO_OFFSET

        # avoid sequence operations for inactive sequencers, including synchronization
        if sequence.is_empty:
            return cls(module=module)

        # conditional configurations
        cfg = cls(
            module=module,
            # connect to physical address
            address=address.local_address,
            # TODO: mixer calibration not yet propagated
            offset_awg_path0=0.0,
            offset_awg_path1=0.0,
            # TODO: properly document - the first 4 marker bits are used to toggle
            # outputs, enabling suitable amplification
            marker_ovr_en=True,
            marker_ovr_value=15,
            # upload sequence
            # - ensure JSON compatibility of the sent dictionary
            sequence=json.loads(sequence.model_dump_json()),
            # configure the sequencers to synchronize
            sync_en=True,
            # modulation, only disable for QCM - always used for flux pulses
            mod_en_awg=rf,
        )

        # acquisition
        if address.input:
            assert isinstance(config, AcquisitionConfig)
            length = _integration_length(sequence)
            if length is not None:
                cfg.integration_length_acq = length
            # discrimination
            if config.iq_angle is not None:
                cfg.thresholded_acq_rotation = np.degrees(config.iq_angle % (2 * np.pi))
            if config.threshold is not None:
                # threshold needs to be compensated by length
                # see: https://docs.qblox.com/en/main/api_reference/sequencer.html#Sequencer.thresholded_acq_threshold
                cfg.thresholded_acq_threshold = config.threshold * length
            # demodulation
            cfg.demod_en_acq = acquisition is not AcquisitionType.RAW

        # set NCO frequency
        # note that probe channels also include readout ones (probe+acquisition), thus
        # there is no need to set it separately for the acquisition (which is on the
        # same IO sequencer)
        probe = channels[channel_id].iqout(channel_id)
        if probe is not None:
            freq = cast(IqConfig, configs[probe]).frequency
            lo = cast(IqChannel, channels[probe]).lo
            assert lo is not None
            lo_freq = cast(OscillatorConfig, configs[lo]).frequency
            cfg.nco_freq = int(freq - lo_freq)

        return cfg

    def apply(self, seq: Sequencer):
        """Configure sequencer-wide settings."""

        mod = cast(Module, seq.ancestors[1])
        for name, value in self.module.items():
            mod.set(name, value)

        if self.address is not None:
            seq.connect_sequencer(self.address)

        # values already applied
        applied = {"module", "address"}
        for name in self.model_fields_set - applied:
            value = getattr(self, name)
            if value is not None:
                seq.set(name, value)



from functools import cached_property
from typing import Union

from .abstract import Operator, OperatorEvolution, SimulationEngine

__all__ = ["QutipEngine"]


class QutipEngine(SimulationEngine):
    """Qutip simulation engine."""

    @cached_property
    def engine(self):
        """Return the qutip engine."""
        # TODO: maybe it can be improved
        import qutip as qt

        return qt

    def evolve(
        self,
        hamiltonian: Operator,
        initial_state: Operator,
        time: list[float],
        time_hamiltonian: OperatorEvolution = None,
        collapse_operators: list[Operator] = None,
        **kwargs,
    ):
        """Evolve the system."""
        if time_hamiltonian is not None:
            hamiltonian += self.engine.QobjEvo(time_hamiltonian.operators)
        return self.engine.mesolve(
            hamiltonian, initial_state, time, collapse_operators, **kwargs
        )

    def create(self, n: int) -> Operator:
        """Create operator for n levels system."""
        return self.engine.create(n)

    def destroy(self, n: int) -> Operator:
        """Destroy operator for n levels system."""
        return self.engine.destroy(n)

    def identity(self, n: int) -> Operator:
        """Identity operator for n levels system."""
        return self.engine.qeye(n)

    def tensor(self, operators: list[Operator]) -> Operator:
        """Tensor product of a list of operators."""
        return self.engine.tensor(*operators)

    def expand(self, op: Operator, targets: Union[int, list[int]], dims: list[int]):
        """Expand operator in larger Hilbert space."""
        return self.engine.expand_operator(op, targets, dims)

    def basis(self, dim: int, state: int) -> Operator:
        """Basis operator for n levels system."""
        return self.engine.basis(dimensions=dim, n=state)



from typing import Annotated, Literal, Optional

from pydantic import Field

from qibolab._core.components import AcquisitionConfig
from qibolab._core.serialize import NdArray

__all__ = ["QcsAcquisitionConfig"]


class QcsAcquisitionConfig(AcquisitionConfig):
    """Acquisition config for Keysight QCS."""

    kind: Literal["qcs-acquisition"] = "qcs-acquisition"

    state_iq_values: Annotated[Optional[NdArray], Field(repr=False)] = None



from functools import reduce
from typing import cast

import numpy as np

from qibolab._core.components import AcquisitionConfig, Configs, OscillatorConfig
from qibolab._core.execution_parameters import AveragingMode, ExecutionParameters
from qibolab._core.identifier import ChannelId, Result
from qibolab._core.pulses import Acquisition, Readout
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import ParallelSweepers

from .validate import ACQUISITION_MEMORY


def batch_shots(
    sequence: PulseSequence,
    sweepers: list[ParallelSweepers],
    options: ExecutionParameters,
) -> list[int]:
    """Subdivide shots in batches.

    It assumes an integrated/discriminated acquisition, such that each
    acquisition in an individual loop accounts for a single bin.
    """
    assert options.nshots is not None

    if options.averaging_mode is not AveragingMode.SINGLESHOT:
        return [options.nshots]

    bins = np.prod(options.bins(sweepers)[1:])
    acquisitions = max(
        sum(1 for p in pulses if isinstance(p, (Acquisition, Readout)))
        for pulses in sequence.by_channel.values()
    )
    per_shot_memory = bins * acquisitions
    max_shots = int(ACQUISITION_MEMORY // per_shot_memory)
    nfull, remainder = np.divmod(options.nshots, max_shots)
    return [max_shots] * int(nfull) + [int(remainder)]


def concat_shots(
    results: list[dict[int, Result]], options: ExecutionParameters
) -> dict[int, Result]:
    """Concatenate shots batches.

    It collects the measurements' batches resulting from the subdivision generated by
    the :func:`batch_shots` function.
    It assumes homogenuous batches, which are only joined over the single outermost
    dimension, and only for :obj:`AveragingMode.SINGLESHOT` acquisitions.
    """
    return (
        reduce(lambda d, e: {k: np.vstack((v, e[k])) for k, v in d.items()}, results)
        if options.averaging_mode is AveragingMode.SINGLESHOT
        else results[0]
    )


def lo_configs(
    los: dict[ChannelId, str], configs: Configs
) -> dict[ChannelId, OscillatorConfig]:
    """Extract local oscillators' configurations."""
    return {ch: cast(OscillatorConfig, configs[lo]) for ch, lo in los.items()}


def time_of_flights(configs: Configs) -> dict[ChannelId, float]:
    """Extract time of flights configurations."""
    return {
        ch: max(config.delay, 4.0)
        for ch, config in configs.items()
        if isinstance(config, AcquisitionConfig)
    }



from functools import cached_property
from typing import Literal, Optional, Union

import numpy as np
from pydantic import Field
from qibo.config import raise_error
from scipy.constants import giga

from ...components import Config
from ...identifier import QubitId, QubitPairId, TransitionId
from ...parameters import Update, _setvalue
from ...pulses import Delay, Pulse, PulseLike, VirtualZ
from ...serialize import Model
from .engine import Operator, SimulationEngine

__all__ = ["DriveEmulatorConfig", "FluxEmulatorConfig", "HamiltonianConfig"]


class DriveEmulatorConfig(Config):
    """Configuration for an IQ channel."""

    kind: Literal["drive-emulator"] = "drive-emulator"

    frequency: float
    """Frequency of drive."""
    rabi_frequency: float = 1e9
    """Rabi frequency [Hz]"""
    scale_factor: float = 1
    """Scaling factor."""

    @staticmethod
    def operator(n: int, engine: SimulationEngine) -> Operator:
        return -1j * (engine.destroy(n) - engine.create(n))


class FluxEmulatorConfig(Config):
    """Configuration for a flux line."""

    kind: Literal["flux-emulator"] = "flux-emulator"

    offset: float
    """DC offset of the channel."""
    voltage_to_flux: float = 1
    """Convert voltarget to flux."""

    @staticmethod
    def operator(n: int, engine: SimulationEngine) -> Operator:
        return engine.create(n) * engine.destroy(n)

    @property
    def flux(self) -> float:
        """Returns flux."""
        return self.offset * self.voltage_to_flux


class Qubit(Config):
    """Hamiltonian parameters for single qubit."""

    frequency: float = 0
    """Qubit frequency for 0->1."""
    anharmonicity: float = 0
    """Qubit anharmonicity."""
    sweetspot: float = 0
    """Sweetspot point."""
    asymmetry: float = 0
    """Asymmetry."""
    t1: dict[TransitionId, float] = Field(default_factory=dict)
    """Dictionary with relaxation times per transition."""
    t2: dict[TransitionId, float] = Field(default_factory=dict)
    """Dictionary with dephasing time per transition."""

    def omega(self, flux: float = 0) -> float:
        """Angular velocity."""
        return 2 * np.pi * self.detuned_frequency(flux)

    def detuned_frequency(self, flux: float) -> float:
        """Return frequency of the qubit modified by the flux."""
        return (self.frequency - self.anharmonicity) * (
            self.asymmetry**2
            + (1 - self.asymmetry**2) * np.cos(np.pi * (flux - self.sweetspot)) ** 2
        ) ** (1 / 4) + self.anharmonicity

    def operator(self, n: int, engine: SimulationEngine, flux: float = 0):
        """Time independent operator."""
        quadratic_term = engine.create(n) * engine.destroy(n) * self.omega(flux) / giga
        quartic_term = (
            self.anharmonicity
            * np.pi
            / giga
            * engine.create(n)
            * engine.create(n)
            * engine.destroy(n)
            * engine.destroy(n)
        )
        return quadratic_term + quartic_term

    def t_phi(self, transition: TransitionId) -> float:
        """T_phi computed from T1 and T2 per transition."""
        return 1 / (1 / self.t2[transition] - 1 / self.t1[transition] / 2)

    def relaxation(self, n: int, engine: SimulationEngine) -> Operator:
        return sum(
            np.sqrt(1 / t1)
            * engine.basis(state=transition[0], dim=n)
            * engine.basis(state=transition[1], dim=n).dag()
            for transition, t1 in self.t1.items()
        )

    def dephasing(self, n: int, engine: SimulationEngine) -> Operator:
        return sum(
            np.sqrt(1 / self.t_phi(pair) / 2)
            * (
                engine.basis(state=pair[0], dim=n)
                * engine.basis(state=pair[0], dim=n).dag()
                - engine.basis(state=pair[1], dim=n)
                * engine.basis(state=pair[1], dim=n).dag()
            )
            for pair in self.t2
        )


class CapacitiveCoupling(Config):
    """Hamiltonian parameters for qubit pair."""

    coupling: float
    """Qubit-qubit coupling."""

    @staticmethod
    def _operator(n: int, engine: SimulationEngine) -> Operator:
        """Time independent operator."""
        op = engine.tensor(
            [
                engine.destroy(n),
                engine.create(n),
            ]
        ) + engine.tensor(
            [
                engine.create(n),
                engine.destroy(n),
            ]
        )
        return 2 * np.pi * op / giga

    def operator(self, n: int, engine: SimulationEngine) -> Operator:
        """Time independent operator."""
        return self.coupling * self._operator(n, engine)


class FluxPulse(Model):
    """Flux pulse term in Hamiltonian."""

    pulse: Pulse
    """Flux pulse to be played."""
    config: FluxEmulatorConfig
    """Flux emulator configuration."""
    qubit: Qubit
    """Qubit affected by the flux pulse."""
    sampling_rate: float = 1
    """Sampling rate."""

    @cached_property
    def envelopes(self):
        """Pulse envelopes."""
        return self.pulse.envelopes(self.sampling_rate)

    @property
    def duration(self):
        """Duration of the pulse."""
        return self.pulse.duration

    @property
    def phase(self):
        """Virtual Z phase."""
        return 0

    def __call__(self, t, sample, phase):
        i, _ = self.envelopes
        # we are passing the relative frequency because the term with the offset
        # is already included in the time-independent part of the Hamiltonian
        # and it corresponds to changing the static bias
        return (
            2
            * np.pi
            * (
                self.qubit.detuned_frequency(
                    self.config.voltage_to_flux * (i[sample] + self.config.offset)
                )
                - self.qubit.detuned_frequency(self.config.flux)
            )
            / giga
        )


class ModulatedDrive(Model):
    """Hamiltonian parameters for qubit drive."""

    pulse: Pulse
    """Drive pulse."""
    config: DriveEmulatorConfig
    """Drive emulator configuration."""
    phase: float = 0
    """Drive has zero virtual z phase."""
    sampling_rate: float = 1
    """Sampling rate."""

    @cached_property
    def envelopes(self):
        """Pulse envelopes."""
        return self.pulse.envelopes(self.sampling_rate)

    @property
    def duration(self):
        """Duration of the pulse."""
        return self.pulse.duration

    @property
    def omega(self):
        return 2 * np.pi * self.config.frequency / giga

    @property
    def rabi_omega(self):
        return 2 * np.pi * self.config.rabi_frequency / giga

    def __call__(self, t, sample, phase):
        i, q = self.envelopes
        phi = self.omega * t + self.pulse.relative_phase + phase
        return (
            self.rabi_omega
            * self.config.scale_factor
            * (np.cos(phi) * i[sample] + np.sin(phi) * q[sample])
        )


class ModulatedDelay(Model):
    """Modulated delay."""

    duration: float
    """Delay duration."""
    phase: float = 0
    """Delay has 0 virtual z phase."""

    def __call__(self, t: float, sample: int, phase: float) -> float:
        """Delay waveform."""
        return 0


class ModulatedVirtualZ(Model):
    """Modulated Virtual Z pulse."""

    phase: float
    """Virtual Z phase."""
    duration: float = 0
    """Duration is 0 for virtual Z."""

    def __call__(self, t: float, sample: int, phase: float) -> float:
        """Delay waveform."""
        raise_error(ValueError, "VirtualZ doesn't have waveform.")


Modulated = Union[ModulatedDrive, ModulatedDelay, ModulatedVirtualZ]
ControlLine = Union[Modulated, FluxPulse]


class HamiltonianConfig(Config):
    """Hamiltonian configuration."""

    kind: Literal["hamiltonian"] = "hamiltonian"
    transmon_levels: int = 2
    qubits: dict[QubitId, Qubit] = Field(default_factory=dict)
    pairs: dict[QubitPairId, CapacitiveCoupling] = Field(default_factory=dict)

    @property
    def nqubits(self):
        """Number of qubits."""
        return len(self.qubits)

    def replace(self, update: Update) -> "HamiltonianConfig":
        """Update parameters' values."""
        d = self.model_dump()
        for path, val in update.items():
            _setvalue(d, path, val)
        return self.model_validate(d)

    def initial_state(self, engine: SimulationEngine) -> Operator:
        """Initial state as ground state of the system."""
        return engine.basis(self.dims, self.nqubits * [0])

    def hilbert_space_index(self, qubit: QubitId) -> int:
        """Return Hilbert space index from qubit id."""
        return list(self.qubits).index(qubit)

    @property
    def dims(self) -> list[int]:
        """Dimensions of the system."""
        return [self.transmon_levels] * len(self.qubits)

    def hamiltonian(self, config: dict, engine: SimulationEngine) -> Operator:
        """Time independent part of Hamiltonian."""
        qubit_terms = sum(
            engine.expand(
                qubit.operator(
                    n=self.transmon_levels,
                    flux=static_flux(qubit=i, config=config),
                    engine=engine,
                ),
                self.dims,
                self.hilbert_space_index(i),
            )
            for i, qubit in self.qubits.items()
        )
        coupling = sum(
            engine.expand(
                pair.operator(self.transmon_levels, engine),
                self.dims,
                [
                    self.hilbert_space_index(pair_id[0]),
                    self.hilbert_space_index(pair_id[1]),
                ],
            )
            for (pair_id, pair) in self.pairs.items()
        )
        return qubit_terms + coupling

    def dissipation(self, engine: SimulationEngine) -> Operator:
        """Dissipation operators for the hamiltonian.

        They are going to be passed to mesolve as collapse operators."""
        collapse_operators = []
        for i, qubit in self.qubits.items():
            if len(qubit.t1) > 0:
                collapse_operators.append(
                    engine.expand(
                        qubit.relaxation(self.transmon_levels, engine),
                        self.dims,
                        self.hilbert_space_index(i),
                    )
                )
            if len(qubit.t2) > 0:
                collapse_operators.append(
                    engine.expand(
                        qubit.dephasing(self.transmon_levels, engine),
                        self.dims,
                        self.hilbert_space_index(i),
                    )
                )
        return collapse_operators


def static_flux(qubit: QubitId, config: dict) -> float:
    """Get static flux for qubit given config (offset)."""
    qubit_config = config.get(f"{qubit}/flux")
    if qubit_config is not None:
        return qubit_config.flux
    coupler_config = config.get(f"coupler_{qubit}/flux")
    if coupler_config is not None:
        return coupler_config.flux
    return 0


def waveform(
    pulse: PulseLike,
    config: Config,
    qubit: Qubit,
) -> Optional[ControlLine]:
    """Convert pulse to hamiltonian."""
    if not isinstance(config, (DriveEmulatorConfig, FluxEmulatorConfig)):
        return None

    if isinstance(pulse, Pulse):
        if isinstance(config, DriveEmulatorConfig):
            return ModulatedDrive(pulse=pulse, config=config)
        if isinstance(config, FluxEmulatorConfig):
            return FluxPulse(
                pulse=pulse,
                config=config,
                qubit=qubit,
            )
    if isinstance(pulse, Delay):
        return ModulatedDelay(duration=pulse.duration)
    if isinstance(pulse, VirtualZ):
        return ModulatedVirtualZ(phase=pulse.phase)



"""Driver for Keysight Quantum Control System (QCS)."""

from qibolab._core.instruments import keysight
from qibolab._core.instruments.keysight import *  # noqa: F403

__all__ = []
__all__ += keysight.__all__



import warnings
from typing import Union

import numpy as np
import numpy.typing as npt
from qm.qua import Cast, Variable


def from_array(var: Variable, array: Union[npt.NDArray, list]):
    """Function parametrizing the QUA `for_` loop from a python array.

    Taken from qualang_tools to avoid the dependency.

    Args:
        var: the QUA variable that will be looped over (int or fixed).
        array: a Python list or numpy array containing the values over which `a` will be looping.
            The spacing must be even in linear or logarithmic scales and it cannot be a QUA array.

    Returns:
        QUA for_ loop parameters (var, init, cond, update) as defined in https://qm-docs.qualang.io/api_references/qua/dsl_main?highlight=for_#qm.qua._dsl.for_.
    """

    # Check for array length
    if len(array) == 0:
        raise Exception("The array must be of length > 0.")
    elif len(array) == 1:
        return var, array[0], var <= array[0], var + 1
    # Check QUA vs python variables
    if not isinstance(var, Variable):
        raise Exception("The first argument must be a QUA variable.")
    if (not isinstance(array[0], (np.generic, int, float))) or (
        isinstance(array[0], bool)
    ):
        raise Exception("The array must be an array of python variables.")
    # Check array increment
    if np.isclose(np.std(np.diff(array)), 0):
        increment = "lin"
    elif np.isclose(np.std(array[1:] / array[:-1]), 0, atol=1e-3):
        increment = "log"
    else:
        raise Exception(
            "The spacing of the input array must be even in linear or logarithmic scales. Please use `for_each_()` for arbitrary scans."
        )

    # Get the type of the specified QUA variable
    start = array[0]
    stop = array[-1]
    # Linear increment
    if increment == "lin":
        step = array[1] - array[0]

        if var.is_int():
            # Check that the array is an array of int with integer increments
            if (
                not float(step).is_integer()
                or not float(start).is_integer()
                or not float(stop).is_integer()
            ):
                raise Exception(
                    "When looping over a QUA int variable, the step and array elements must be integers."
                )
            # Generate the loop parameters for positive and negative steps
            if step > 0:
                return var, int(start), var <= int(stop), var + int(step)
            else:
                return var, int(start), var >= int(stop), var + int(step)

        elif var.is_fixed():
            # Check for fixed number overflows
            if not (-8 <= start < 8) and not (-8 <= stop < 8):
                raise Exception("fixed numbers are bounded to [-8, 8).")

            # Generate the loop parameters for positive and negative steps
            if step > 0:
                return (
                    var,
                    float(start),
                    var < float(stop) + float(step) / 2,
                    var + float(step),
                )
            else:
                return (
                    var,
                    float(start),
                    var > float(stop) + float(step) / 2,
                    var + float(step),
                )
        else:
            raise Exception(
                "This variable type is not supported. Please use a QUA 'int' or 'fixed' or contact a QM member for assistance."
            )
    # Logarithmic increment
    elif increment == "log":
        step = array[1] / array[0]

        if var.is_int():
            warnings.warn(
                "When using logarithmic increments with QUA integers, the resulting values will slightly differ from the ones in numpy.logspace() because of rounding errors. \n Please use the get_equivalent_log_array() function to get the exact values taken by the QUA variable and note that the number of points may also change."
            )
            if step > 1:
                if int(round(start) * float(step)) == int(round(start)):
                    raise ValueError(
                        "Two successive values in the scan are equal after being cast to integers which will make the QUA for_ loop fail. \nEither increase the logarithmic step or use for_each_(): https://docs.quantum-machines.co/1.1.6/qm-qua-sdk/docs/Guides/features/?h=for_ea#for_each."
                    )
                else:
                    return (
                        var,
                        round(start),
                        var < round(stop) * np.sqrt(float(step)),
                        Cast.mul_int_by_fixed(var, float(step)),
                    )
            else:
                return (
                    var,
                    round(start),
                    var > round(stop) / np.sqrt(float(step)),
                    Cast.mul_int_by_fixed(var, float(step)),
                )

        elif var.is_fixed():
            # Check for fixed number overflows
            if not (-8 <= start < 8) and not (-8 <= stop < 8):
                raise Exception("fixed numbers are bounded to [-8, 8).")

            if step > 1:
                return (
                    var,
                    float(start),
                    var < float(stop) * np.sqrt(float(step)),
                    var * float(step),
                )
            else:
                return (
                    var,
                    float(start),
                    var > float(stop) * np.sqrt(float(step)),
                    var * float(step),
                )
        else:
            raise Exception(
                "This variable type is not supported. Please use a QUA 'int' or 'fixed' or contact a QM member for assistance."
            )



from collections.abc import Callable, Iterable
from enum import Enum, auto
from itertools import groupby
from typing import Optional

from qibolab._core.identifier import ChannelId
from qibolab._core.instruments.qblox.q1asm.ast_ import SetAwgGain, SetAwgOffs, SetFreq
from qibolab._core.pulses.pulse import (
    Pulse,
    PulseId,
    PulseLike,
)
from qibolab._core.serialize import Model
from qibolab._core.sweeper import ParallelSweepers, Parameter, Range, Sweeper

from ..q1asm.ast_ import (
    Instruction,
    Register,
    SetPhDelta,
    Value,
)
from .asm import MAX_PARAM, convert

__all__ = []


class ParamRole(Enum):
    """Parameter role in sweep.

    Discriminate the various register roles. Particularly, in the pulse
    duration sweeps.
    """

    FREQUENCY = auto(), Parameter.frequency
    "Channel frequency."
    OFFSET = auto(), Parameter.offset
    "Channel offset."
    AMPLITUDE = auto(), Parameter.amplitude
    "Pulse amplitude."
    PHASE = auto(), Parameter.relative_phase
    "Pulse relative phase."
    DURATION = auto(), Parameter.duration
    "Pulse duration."
    PULSE_I = auto(), Parameter.duration
    "Pulse I component."
    PULSE_Q = auto(), Parameter.duration
    "Pulse Q component."

    @classmethod
    def from_sweeper(cls, sweep: Sweeper) -> "ParamRole":
        for var in cls:
            if sweep.parameter is var.kind:
                return var
        raise ValueError("Sweeper parameter kind does not correspond to any role.")

    @classmethod
    def unique(cls, sweep: Sweeper) -> bool:
        return sweep.parameter is not Parameter.duration or (
            sweep.pulses is not None
            and not any(isinstance(p, Pulse) for p in sweep.pulses)
        )

    @property
    def kind(self):
        return self.value[1]


class Param(Model):
    reg: Register
    """Register used for the parameter value."""
    start: int
    """Initial value."""
    step: int
    """Increment."""
    role: ParamRole
    """The parameter type."""
    pulse: Optional[PulseId]
    """The target pulse (if the sweeper targets pulses)."""
    channel: Optional[ChannelId]
    """The target channel (if the sweeper targets channels)."""
    loop: Optional[int] = None
    """The loop which is associated to."""

    @property
    def description(self) -> str:
        """Textual description, used in some accompanying comments."""
        return (
            "sweeper ("
            + (f"loop: {self.loop}, " if self.loop is not None else "")
            + (f"pulse: 0x{self.pulse.hex[:5]}, " if self.pulse is not None else "")
            + f"role: {self.role.kind.name}.{self.role.name}"
            + ")"
        )


class LoopParams(Model):
    """Parameters involved in a single loop level."""

    channel: list[Param]
    pulse: list[Param]

    @property
    def all(self):
        return self.channel + self.pulse


IndexedParams = dict[int, LoopParams]
"""Sweep parameters, organized by loop.

Keys are going to be loop counters.
"""


def _channels_pulses(
    pars: Iterable[Param],
) -> LoopParams:
    channels = []
    pulses = []
    for p in pars:
        (channels if p.channel is not None else pulses).append(p)
    return LoopParams(channel=channels, pulse=pulses)


def params_reshape(params: list[Param]) -> IndexedParams:
    """Split parameters related to channels and pulses.

    Moreover, it reorganize them by loop, to group the updates.
    """
    return {
        key: _channels_pulses(pars)
        for key, pars in groupby(params, key=lambda p: p.loop)
        if key is not None
    }


def _pulse_duration(sweep: Sweeper) -> list[tuple[Range, "ParamRole"]]:
    """Reserve 3 registers for a pulse duration sweeper."""
    return [
        ((0, 2 * len(sweep), 2), ParamRole.PULSE_I),
        ((1, 2 * len(sweep) + 1, 2), ParamRole.PULSE_Q),
        ((sweep - 4.0).irange, ParamRole.DURATION),
    ]


def _registers(sweep: Sweeper) -> list[tuple[Range, ParamRole]]:
    """Reserve registers for sweeping."""
    return (
        [(sweep.irange, ParamRole.from_sweeper(sweep))]
        if ParamRole.unique(sweep)
        else _pulse_duration(sweep)
    )


def _unravel_sweeps(sweepers: list[ParallelSweepers]) -> Iterable[tuple[int, Param]]:
    """Turn sweepers into suitable ranges with unique targets."""
    return (
        (
            j,
            Param(
                reg=Register(number=0),
                start=int(convert(irange[0], sweep.parameter)),
                step=int(convert(irange[2], sweep.parameter)),
                pulse=pulse.id if pulse is not None else None,
                channel=channel,
                role=role,
            ),
        )
        # the first sweeper should be the outermost, thus reverse them during the
        # enumeration
        for j, parsweep in enumerate(sweepers[::-1])
        for sweep in parsweep
        for irange, role in _registers(sweep)
        for pulse in (sweep.pulses if sweep.pulses is not None else [None])
        for channel in (sweep.channels if sweep.channels is not None else [None])
    )


def params(sweepers: list[ParallelSweepers], allocated: int) -> list[Param]:
    """Initialize parameters' registers.

    `allocated` is the number of already allocated registers for loop counters, as
    initialized by :func:`loops`.
    """
    return [
        p.model_copy(update={"reg": Register(number=i + allocated + 1), "loop": j})
        for i, (j, p) in enumerate(_unravel_sweeps(sweepers))
    ]


class _Update(Model):
    update: Optional[Callable[[Value], Instruction]]
    reset: Optional[Callable[[Value], Instruction]]


_SWEEP_UPDATE: dict[Parameter, _Update] = {
    Parameter.frequency: _Update(update=lambda v: SetFreq(value=v), reset=None),
    Parameter.offset: _Update(
        update=lambda v: SetAwgOffs(value_0=v, value_1=v), reset=None
    ),
    Parameter.amplitude: _Update(
        update=lambda v: SetAwgGain(value_0=v, value_1=v),
        reset=lambda _: SetAwgGain(
            value_0=MAX_PARAM[Parameter.amplitude],
            value_1=MAX_PARAM[Parameter.amplitude],
        ),
    ),
    Parameter.relative_phase: _Update(update=lambda v: SetPhDelta(value=v), reset=None),
    Parameter.duration: _Update(update=None, reset=None),
}


def update_instructions(
    role: ParamRole, value: Value, reset: bool = False
) -> list[Instruction]:
    wrapper = _SWEEP_UPDATE[role.kind]
    up = wrapper.update if not reset else wrapper.reset
    instr = [up(value)] if up is not None else []
    return instr


def reset_instructions(role: ParamRole, value: Value) -> list[Instruction]:
    return update_instructions(role, value, reset=True)


ParameterizedPulse = tuple[PulseLike, set[Param]]
SweepSequence = list[ParameterizedPulse]


def sweep_sequence(sequence: Iterable[PulseLike], params: list[Param]) -> SweepSequence:
    """Wrap swept pulses with updates markers."""
    parsbyid = {
        p: {pair[1] for pair in pairs}
        for p, pairs in groupby(
            sorted(
                ((p.pulse, p) for p in params if p.pulse is not None),
                key=lambda pair: pair[0],
            ),
            key=lambda pair: pair[0],
        )
    }
    return [(e, parsbyid.get(e.id, set())) for e in sequence]



from dataclasses import asdict, dataclass, field
from typing import Optional, Union

import numpy as np

from qibolab._core.components import (
    AcquisitionChannel,
    DcChannel,
    IqChannel,
    IqConfig,
    OscillatorConfig,
)
from qibolab._core.identifier import ChannelId
from qibolab._core.pulses import Custom, Pulse, Readout

from ..components import MwFemOscillatorConfig, OpxOutputConfig, QmAcquisitionConfig
from .devices import (
    Controller,
    ControllerId,
    Controllers,
    ModuleTypes,
    MwFemInput,
    MwFemOutput,
    Octave,
    OctaveInput,
    OctaveOutput,
)
from .elements import (
    AcquireMwFemElement,
    AcquireOctaveElement,
    DcElement,
    Element,
    MwFemElement,
    RfOctaveElement,
)
from .pulses import (
    QmAcquisition,
    QmPulse,
    Waveform,
    integration_weights,
    operation,
    waveforms_from_pulse,
)

__all__ = ["Configuration"]

DEFAULT_DIGITAL_WAVEFORMS = {"ON": {"samples": [(1, 0)]}}
"""Required to be registered in the config for QM to work.

Also used as triggering that allows the Octave LO signal to pass only
when we are executing something.
"""


@dataclass
class Configuration:
    """Configuration for communicating with the ``QuantumMachinesManager``.

    Contains nested ``dataclass`` objects and is serialized using ``asdict``
    to be sent to the instrument.
    """

    version: int = 1
    controllers: Controllers = field(default_factory=Controllers)
    octaves: dict[str, Octave] = field(default_factory=dict)
    elements: dict[str, Element] = field(default_factory=dict)
    pulses: dict[str, Union[QmPulse, QmAcquisition]] = field(default_factory=dict)
    waveforms: dict[str, Waveform] = field(default_factory=dict)
    digital_waveforms: dict = field(
        default_factory=lambda: DEFAULT_DIGITAL_WAVEFORMS.copy()
    )
    integration_weights: dict = field(default_factory=dict)
    mixers: dict = field(default_factory=dict)

    def add_controller(self, device: ControllerId, modules: dict[str, ModuleTypes]):
        if device not in self.controllers:
            self.controllers[device] = Controller(type=modules[device])

    def add_octave(
        self, device: str, connectivity: ControllerId, modules: dict[str, ModuleTypes]
    ):
        if device not in self.octaves:
            self.add_controller(connectivity, modules)
            self.octaves[device] = Octave(connectivity)

    def configure_dc_line(
        self, id: ChannelId, channel: DcChannel, config: OpxOutputConfig
    ):
        controller = self.controllers[channel.device]
        if controller.type == "opx1":
            keys = ["offset"]
        else:
            keys = list(config.model_fields.keys())
            keys.remove("kind")
        config_values = config.model_dump()
        values = {k: config_values[k] for k in keys}
        values.update({"filter": config.filter})
        if config.sampling_rate > 1e9:
            del values["upsampling_mode"]
        controller.analog_outputs[channel.port] = values
        self.elements[id] = DcElement.from_channel(channel)

    def configure_mw_fem_line(
        self,
        channel: IqChannel,
        config: IqConfig,
        lo_config: MwFemOscillatorConfig,
        id: Optional[ChannelId] = None,
    ):
        controller = self.controllers[channel.device]
        if channel.port in controller.analog_outputs:
            output = MwFemOutput(**controller.analog_outputs[channel.port])
            output.update(lo_config)
        else:
            output = MwFemOutput.from_config(lo_config)
        controller.analog_outputs[channel.port] = asdict(output)
        if id is not None:
            intermediate_frequency = config.frequency - lo_config.frequency
            self.elements[id] = MwFemElement.from_channel(
                channel, lo_config.upconverter, intermediate_frequency
            )

    def configure_iq_line(
        self,
        channel: IqChannel,
        config: IqConfig,
        lo_config: OscillatorConfig,
        id: Optional[ChannelId] = None,
    ):
        port = channel.port
        octave = self.octaves[channel.device]
        octave.RF_outputs[port] = OctaveOutput.from_config(lo_config)
        self.controllers[octave.connectivity].add_octave_output(port)

        if id is not None:
            intermediate_frequency = config.frequency - lo_config.frequency
            self.elements[id] = RfOctaveElement.from_channel(
                channel, octave.connectivity, intermediate_frequency
            )

    def configure_mw_fem_acquire_line(
        self,
        acquire_channel: AcquisitionChannel,
        probe_channel: IqChannel,
        acquire_config: QmAcquisitionConfig,
        probe_config: IqConfig,
        lo_config: MwFemOscillatorConfig,
        id: ChannelId,
    ):
        port = acquire_channel.port
        controller = self.controllers[acquire_channel.device]
        controller.analog_inputs[port] = MwFemInput.from_config(lo_config)

        self.configure_mw_fem_line(probe_channel, probe_config, lo_config)

        intermediate_frequency = probe_config.frequency - lo_config.frequency
        self.elements[id] = AcquireMwFemElement.from_channel(
            probe_channel,
            lo_config.upconverter,
            acquire_channel,
            intermediate_frequency=intermediate_frequency,
            time_of_flight=acquire_config.delay,
            smearing=acquire_config.smearing,
        )

    def configure_acquire_line(
        self,
        acquire_channel: AcquisitionChannel,
        probe_channel: IqChannel,
        acquire_config: QmAcquisitionConfig,
        probe_config: IqConfig,
        lo_config: OscillatorConfig,
        id: ChannelId,
    ):
        port = acquire_channel.port
        octave = self.octaves[acquire_channel.device]
        octave.RF_inputs[port] = OctaveInput(lo_config.frequency)
        self.controllers[octave.connectivity].add_octave_input(port, acquire_config)

        self.configure_iq_line(probe_channel, probe_config, lo_config)

        intermediate_frequency = probe_config.frequency - lo_config.frequency
        self.elements[id] = AcquireOctaveElement.from_channel(
            probe_channel,
            acquire_channel,
            octave.connectivity,
            intermediate_frequency,
            time_of_flight=acquire_config.delay,
            smearing=acquire_config.smearing,
        )

    def register_waveforms(
        self,
        pulse: Pulse,
        sampling_rate: int,
        max_voltage: float,
        element: Optional[str] = None,
        dc: bool = False,
    ):
        if dc:
            qmpulse = QmPulse.from_dc_pulse(pulse, sampling_rate)
        else:
            if element is None:
                qmpulse = QmPulse.from_pulse(pulse, sampling_rate)
            else:
                qmpulse = QmAcquisition.from_pulse(pulse, element)
        waveforms = waveforms_from_pulse(pulse, sampling_rate, max_voltage)
        if dc:
            self.waveforms[qmpulse.waveforms["single"]] = waveforms["I"]
        else:
            for mode in ["I", "Q"]:
                self.waveforms[getattr(qmpulse.waveforms, mode)] = waveforms[mode]
        return qmpulse

    def register_iq_pulse(
        self, element: str, pulse: Pulse, sampling_rate: int, max_voltage: float
    ):
        op = operation(pulse)
        if op not in self.pulses:
            self.pulses[op] = self.register_waveforms(pulse, sampling_rate, max_voltage)
        self.elements[element].operations[op] = op
        return op

    def register_dc_pulse(
        self, element: str, pulse: Pulse, sampling_rate: int, max_voltage: float
    ):
        op = operation(pulse)
        if op not in self.pulses:
            self.pulses[op] = self.register_waveforms(
                pulse, sampling_rate, max_voltage, dc=True
            )
        self.elements[element].operations[op] = op
        return op

    def register_acquisition_pulse(
        self, element: str, readout: Readout, sampling_rate: int, max_voltage: float
    ):
        """Registers pulse, waveforms and integration weights in QM config."""
        op = operation(readout)
        acquisition = f"{op}_{element}"
        if acquisition not in self.pulses:
            new_probe = readout.probe.model_copy(
                update=dict(
                    duration=readout.acquisition.duration,
                    envelope=Custom(
                        i_=np.pad(
                            readout.probe.envelope.i(int(readout.probe.duration)),
                            (
                                0,
                                int(
                                    readout.acquisition.duration
                                    - readout.probe.duration
                                ),
                            ),
                            mode="constant",
                            constant_values=0,
                        ),
                        q_=np.pad(
                            readout.probe.envelope.q(int(readout.probe.duration)),
                            (
                                0,
                                int(
                                    readout.acquisition.duration
                                    - readout.probe.duration
                                ),
                            ),
                            mode="constant",
                            constant_values=0,
                        ),
                    ),
                )
            )
            self.pulses[acquisition] = self.register_waveforms(
                new_probe,
                sampling_rate,
                max_voltage,
                element,
            )
        self.elements[element].operations[op] = acquisition
        return op

    def register_integration_weights(self, element: str, duration: int, kernel):
        self.integration_weights.update(integration_weights(element, duration, kernel))



from pathlib import Path

from lark import Lark, Transformer

from .ast_ import INSTRUCTIONS, Comment, Line, Program

__all__ = ["parse"]

GRAMMAR_FILE = Path(__file__).parent / "q1asm.lark"
GRAMMAR = GRAMMAR_FILE.read_text()


parser = Lark(GRAMMAR)


class ToAst(Transformer):
    def instruction(self, args):
        name = args[0].data.value
        attrs = (a.value for a in args[0].children)
        return INSTRUCTIONS[name].from_args(*attrs)

    def line(self, args):
        label = args[0].value if args[0] is not None else None
        comment = args[2].value[1:] if args[2] is not None else None
        return Line(instruction=args[1], label=label, comment=comment)

    def comment(self, args):
        return Comment(args[0].value[1:].strip())


def parse(code: str) -> Program:
    """Parse Q1ASM representation."""
    tree = parser.parse(code)
    return Program.from_elements(ToAst().transform(tree).children)



from enum import Enum

import numpy as np

from qibolab._core.sweeper import Parameter

from ..q1asm.ast_ import Instruction, Line, Lineable, Register

__all__ = []


class Registers(Enum):
    """Pre-assigned register numbers."""

    bin = Register(number=0)
    bin_reset = Register(number=1)
    shots = Register(number=2)
    wait = Register(number=3)

    @classmethod
    def first_available(cls) -> int:
        return max(r.value.number for r in cls) + 1


def label(line: Lineable, label: str) -> Line:
    return (
        Line(instruction=line, label=label)
        if isinstance(line, Instruction)
        else Line(instruction=line.instruction, comment=line.comment, label=label)
    )


MAX_PARAM = {
    Parameter.amplitude: 2**15 - 1,
    Parameter.offset: 2**15 - 1,
    Parameter.relative_phase: 1e9,
    Parameter.frequency: 2e9,
}
"""Maximum range for parameters.

Declared in https://docs.qblox.com/en/main/cluster/q1_sequence_processor.html#q1-instructions

Ranges may be one-sided (just positive) or two-sided. This is accounted for in
:func:`convert`.
"""


def _convert_offset(offset: float) -> float:
    """Converts offset values to the encoding used in qblox FPGAs."""

    # TODO: move validation closer to user input
    if abs(offset) >= 1:
        raise ValueError("Offset must be a float between -1 and 1.")

    return np.floor(offset * MAX_PARAM[Parameter.offset])


def convert(value: float, kind: Parameter) -> float:
    """Convert sweeper value in assembly units."""
    if kind is Parameter.amplitude:
        return value * MAX_PARAM[kind]
    if kind is Parameter.relative_phase:
        # TODO: the following is actually redundant, choose what to keep
        # most likely the maximum value, set to 1e9, is something like 2**30 (not sure
        # why not 2**32), and the three % operations are all doing the same
        return ((value % (2 * np.pi)) / (2 * np.pi)) % 1.0 * MAX_PARAM[kind] % (2**32)
    if kind is Parameter.frequency:
        return 4 * value % (2**32)
    if kind is Parameter.offset:
        return _convert_offset(value) % (2**32)
    if kind is Parameter.duration:
        return value
    raise ValueError(f"Unsupported sweeper: {kind.name}")



from functools import reduce
from typing import Optional, cast

from pydantic import BaseModel, Field

from ..q1asm.ast_ import (
    Block,
    Instruction,
    Line,
    Loop,
    Move,
    Program,
    Reference,
    Register,
    Sub,
    Wait,
)
from .asm import Registers

__all__ = []

MAX_WAIT = 2**16 - 1


class State(BaseModel):
    nwait: int = 0


def _long_wait(duration: int, n: int) -> Block:
    """Split a statically long wait.

    It accounts for the wait instruction limit, defined by :const:`MAX_WAIT`.

    ``n`` is used for labelling the loop, and it should be different for each wait
    instruction in a sequencer.
    """
    iterations = duration // MAX_WAIT
    remainder = duration % MAX_WAIT
    register = Registers.wait.value
    label = f"wait{n}"
    return [Wait(duration=remainder)] + [
        Move(source=iterations, destination=register),
        Line(instruction=Wait(duration=MAX_WAIT), label=label),
        Line.instr(Loop(a=register, address=Reference(label=label))),
    ]


def _decompose_wait(instr: Wait, n: int) -> Optional[Block]:
    duration = instr.duration
    if not isinstance(duration, int) or duration <= MAX_WAIT:
        return None
    return _long_wait(duration, n)


def _negative_move(instr: Move):
    """Compile negative value sets.

    Apparently, the only place where negative numbers are not allowed
    are registers, otherwise they are handled by the internal compiler.

    https://docs.qblox.com/en/main/tutorials/q1asm_tutorials/intermediate/nco_control_adv.html#:~:text=Internally,%20the%20processor%20stores

    Thus, we compile instructions setting negative values as suggested:
    first setting them to 0, than subtracting the desired amount. This
    is more reliable than manually complementing the number, since it
    makes no assumption about the registers size.

    https://docs.qblox.com/en/main/cluster/troubleshooting.html#:~:text=How%20do%20I%20set%20negative%20numbers
    """
    src = cast(int, instr.source)
    dest = instr.destination
    return [Move(source=0, destination=dest), Sub(a=dest, b=abs(src), destination=dest)]


def _decompose_move(instr: Move) -> Optional[Block]:
    src = instr.source
    if isinstance(src, Register):
        return None
    assert isinstance(src, int)
    if src >= 0:
        return None
    return _negative_move(instr)


LineTransformed = list[Line]


def _line_transform(line: Line, state: State) -> tuple[LineTransformed, State]:
    instr = line.instruction
    block, state = (
        (
            _decompose_wait(instr, state.nwait),
            state.model_copy(update={"nwait": state.nwait + 1}),
        )
        if isinstance(instr, Wait)
        else (_decompose_move(instr), state)
        if isinstance(instr, Move)
        else (None, state)
    )

    # default
    if block is None:
        return [line], state

    assert isinstance(block[0], Instruction)
    return [
        el if isinstance(el, Line) else Line.instr(el)
        for el in (
            (
                Line(instruction=block[0], label=line.label, comment=line.comment),
                *(el for el in block[1:]),
            )
            if block is not None
            else [line]
        )
    ], state


def _line_transform_apply(
    value: tuple[list[LineTransformed], State], line: Line
) -> tuple[list[LineTransformed], State]:
    """Accumulate."""
    transformed, state = _line_transform(line, value[1])
    return (value[0] + [transformed]), state


class _WaitBatch(BaseModel):
    duration: int = 0
    comment: list[str] = Field(default_factory=list)
    label: Optional[str] = None

    def increment(self, line: Line):
        instr = line.instruction
        assert isinstance(instr, Wait) and isinstance(instr.duration, int)
        self.duration += instr.duration
        if line.comment is not None:
            self.comment.append(line.comment)
        if line.label is not None:
            self.label = line.label

    @property
    def lines(self) -> list[Line]:
        return (
            [
                Line(
                    label=self.label,
                    instruction=Wait(duration=self.duration),
                    comment="\n".join(self.comment),
                )
            ]
            if self.duration > 0
            else []
        )


def _merge_wait(block: list[Line]) -> list[Line]:
    """Merge subsequent static (immediate) waits."""
    batch = _WaitBatch()
    new = []
    for line in block:
        instr = line.instruction
        intwait = isinstance(instr, Wait) and isinstance(instr.duration, int)
        if not intwait or line.label is not None:
            new += batch.lines + ([line] if not intwait else [])
            batch = _WaitBatch()
        if intwait:
            batch.increment(line)

    return new + batch.lines


def _block_transform(block: Block) -> list[Line]:
    lines = [el if isinstance(el, Line) else Line.instr(el) for el in block]
    return _merge_wait(lines)


def transpile(prog: Block) -> Program:
    block_replaced = _block_transform(prog)
    lines_replaced, _state = reduce(
        _line_transform_apply, block_replaced, ([], State())
    )
    return Program(elements=[line for block in lines_replaced for line in block])
